from __future__ import annotations

import pathlib
import shutil

import pytest
from pytest_mock import MockerFixture

from ..main import load_renames, main
from .conftest import LATTICE_FILES

lattice_file = pytest.mark.parametrize(
    "lattice_file", [pytest.param(p, id=p.name) for p in LATTICE_FILES]
)


@pytest.fixture
def input_filename(tmp_path: pathlib.Path, lattice_file: pathlib.Path) -> pathlib.Path:
    dest_file = tmp_path / lattice_file.name
    shutil.copy(lattice_file, dest_file)
    return dest_file


def test_stdin_processing(capsys: pytest.CaptureFixture, mocker: MockerFixture):
    input_content = "d1    : drift   , L   =1.0;"
    mocker.patch("sys.stdin.read", return_value=input_content)

    main(filename="-")

    captured = capsys.readouterr()
    assert "d1: drift, L=1.0" in captured.out.splitlines()


@lattice_file
def test_standard_formatting_stdout(input_filename: pathlib.Path, capsys: pytest.CaptureFixture):
    main(filename=input_filename)

    captured = capsys.readouterr()
    assert len(captured.out) > 0
    assert ":" in captured.out


@lattice_file
def test_formatting_output_file(input_filename: pathlib.Path, tmp_path: pathlib.Path):
    output_file = tmp_path / "_formatted_out.bmad"

    main(filename=input_filename, output=output_file)

    assert output_file.exists()
    assert output_file.stat().st_size > 0


@lattice_file
def test_in_place_modification(input_filename: pathlib.Path):
    original_content = input_filename.read_text()
    main(filename=input_filename, in_place=True)

    assert input_filename.exists()
    new_content = input_filename.read_text()
    assert len(new_content) > 0
    assert new_content != original_content


@lattice_file
def test_in_place_modification_recursive(lattice_file: pathlib.Path, tmp_path: pathlib.Path):
    shutil.copytree(lattice_file.parent, tmp_path, dirs_exist_ok=True)

    input_filename = tmp_path / lattice_file.name
    original_content = input_filename.read_text()
    main(filename=input_filename, in_place=True, recursive=True)

    assert input_filename.exists()
    new_content = input_filename.read_text()
    assert len(new_content) > 0
    assert new_content != original_content


@lattice_file
def test_diff_generation(input_filename: pathlib.Path, capsys: pytest.CaptureFixture):
    main(filename=input_filename, diff=True)

    captured = capsys.readouterr()
    assert "---" in captured.out
    assert "+++" in captured.out


def test_rename_logic(tmp_path: pathlib.Path):
    content = "QF: QUAD, L=1, K1=0.5;"
    f = tmp_path / "test.bmad"
    f.write_text(content)

    main(filename=f, raw_renames=["QF, QUAD_ABC"], compact=True)


def test_rename_functionality(capsys: pytest.CaptureFixture, tmp_path):
    f = tmp_path / "rename_test.bmad"
    f.write_text("OLD_NAME: DRIFT, L=1;")

    main(filename=f, raw_renames=["OLD_NAME, NEW_NAME"])

    captured = capsys.readouterr()
    assert "NEW_NAME: DRIFT, L=1" in captured.out


@lattice_file
def test_verbosity_levels(input_filename: pathlib.Path, capsys: pytest.CaptureFixture):
    main(filename=input_filename, verbose=2)
    captured = capsys.readouterr()
    assert "-- Block" in captured.err


def test_load_renames(tmp_path: pathlib.Path):
    rename_file = tmp_path / "renames.csv"
    rename_file.write_text("A,B\nC,D")

    renames_arg = {"E": "F"}
    raw_renames = ["G,H"]

    result = load_renames(rename_file, raw_renames, renames_arg)

    expected = {"A": "B", "C": "D", "G": "H", "E": "F"}
    assert result == expected



from __future__ import annotations

import enum
import typing
from typing import Any, Sequence, SupportsIndex

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

from .comments import Comments
from .location import Location

if typing.TYPE_CHECKING:
    from .types import Block


class Role(str, enum.Enum):
    name_ = "name"
    builtin = "builtin"
    kind = "kind"
    attribute_name = "attribute_name"
    statement_definition = "statement_definition"
    filename = "filename"


class Token(str):
    """
    String with source loc information.

    Comparisons are case-insensitive.
    """

    loc: Location
    comments: Comments
    role: Role | None = None

    def __new__(
        cls,
        content: str,
        loc: Location | None = None,
        comments: Comments | None = None,
        role: Role | None = None,
    ):
        return super().__new__(cls, content)

    def __init__(
        self,
        content: str,
        loc: Location | None = None,
        comments: Comments | None = None,
        role: Role | None = None,
    ):
        self.loc = loc or Location(end_column=len(content))
        self.comments = comments or Comments()
        self.role = role

        # internal error
        if not isinstance(self.loc, Location):
            raise ValueError(type(self.loc))
        if not isinstance(self.comments, Comments):
            raise ValueError(type(self.comments))
        if role and not isinstance(self.role, Role):
            raise ValueError(type(self.role))

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other) -> bool:
        case_insensitive_equality = str(other).upper() == str(self).upper()
        if hasattr(other, "comments"):
            return case_insensitive_equality and self.comments == other.comments

        return case_insensitive_equality

    def __ne__(self, other) -> bool:
        return not (self == other)

    @property
    def is_quoted_string(self) -> bool:
        return (self.startswith(_SQUOTE) and self.endswith(_SQUOTE)) or (
            self.startswith(_DQUOTE) and self.endswith(_DQUOTE)
        )

    def remove_quotes(self) -> Token:
        if self.is_quoted_string:
            # TODO fix location
            return Token(self[1:-1], loc=self.loc, comments=self.comments)
        return self

    def __repr__(self) -> str:
        # return super().__repr__()
        # parts = [super().__repr__(), self.loc]
        parts: list[Any] = [super().__repr__()]
        if bool(self.comments):
            parts.append(self.comments)
        desc = ", ".join(str(part) for part in parts if part)
        return f"{type(self).__name__}({desc})"

    def annotate(self, named: dict[Token, Any]):
        if self.upper() in named:
            self.role = Role.name_

    @classmethod
    def join(
        cls: type[Token],
        args: Sequence[Token | Block],
        delim: Delimiter | None = None,
        role: Role | None = None,
    ):
        if not args:
            return cls("", role=role)

        from .types import _flatten_blocks

        strs = _flatten_blocks(list(args))
        if not strs:
            return cls("", role=role)

        str0, *_ = strs
        if not str0.loc:
            return cls("".join(str(s) for s in strs), role=role)

        result_parts = [str(str0)]

        if delim and len(strs) > 1:
            result_parts.append(str(delim))
            result_parts.append(" ")

        prev_end_line = str0.loc.end_line or str0.loc.line
        prev_end_col = str0.loc.end_column or (str0.loc.column + len(str(str0)))

        for idx, arg in enumerate(strs[1:], start=2):
            if not arg.loc:
                result_parts.append(str(arg))
                if delim and idx < len(strs):
                    result_parts.append(str(delim))
                    result_parts.append(" ")
                continue

            # Different line: add a space
            if arg.loc.line != prev_end_line:
                result_parts.append(" ")
            # Same line but column is beyond where previous token ended: add a space
            elif arg.loc.column > prev_end_col:
                result_parts.append(" ")

            result_parts.append(str(arg))
            if delim and idx < len(strs):
                result_parts.append(str(delim))
                result_parts.append(" ")

            prev_end_line = arg.loc.end_line or arg.loc.line
            prev_end_col = arg.loc.end_column or (arg.loc.column + len(str(arg)))

        return cls(
            "".join(result_parts),
            loc=Location.from_items(strs),
            comments=str0.comments,
            role=role,
        )

    def quoted(self) -> Self:
        if self.is_quoted_string:
            return self

        value = str(self).strip()
        if '"' in value:
            quote_char = "'"
        else:
            quote_char = '"'
        if quote_char in value:
            # Sorry, but why use both quote types and not quote your string to begin with?
            # I don't think there's an escape character we can use
            value = value.replace(quote_char, " ")

        return type(self)(
            f"{quote_char}{value}{quote_char}",
            loc=self.loc,
            comments=self.comments,
            role=self.role,
        )

    def strip(self, chars: str | None = None) -> Self:
        return type(self)(
            str(self).strip(chars),
            loc=self.loc,
            comments=self.comments,
            role=self.role,
        )

    def replace(self, old, new, count: SupportsIndex = -1) -> Self:
        return type(self)(
            str(self).replace(old, new, count),
            loc=self.loc,
            comments=self.comments,
            role=self.role,
        )

    def with_(
        self,
        *,
        loc: Location | None = None,
        comments: Comments | None = None,
        role: Role | None = None,
    ):
        return type(self)(
            str(self),
            loc=loc or self.loc,
            comments=comments or self.comments,
            role=role or self.role,
        )

    def upper(self):
        return type(self)(
            super().upper(),
            loc=self.loc,
            comments=self.comments,
            role=self.role,
        )

    def lower(self):
        return type(self)(
            super().lower(),
            loc=self.loc,
            comments=self.comments,
            role=self.role,
        )


class Delimiter(Token):
    pass


_DQUOTE = Delimiter('"')
_SQUOTE = Delimiter("'")



import re
from dataclasses import dataclass

from .const import (
    m_deuteron,
    m_electron,
    m_helion,
    m_muon,
    m_neutron,
    m_pion_0,
    m_pion_charged,
    m_proton,
)
from .exceptions import InvalidChargeError, ParticleSpeciesError

# TODO add to SubatomicParticle
# anomalous_moment_of_subatomi = [
#     "anomalous_mag_moment_He3",
#     0.0,
#     "anomalous_mag_moment_neutron",
#     "anomalous_mag_moment_deuteron",
#     0.0,
#     "anomalous_mag_moment_muon",
#     "anomalous_mag_moment_proton",
#     "anomalous_mag_moment_electron",
#     0.0,
#     "anomalous_mag_moment_electron",
#     "anomalous_mag_moment_proton",
#     "anomalous_mag_moment_muon",
#     0.0,
#     "anomalous_mag_moment_deuteron",
#     "anomalous_mag_moment_neutron",
#     0.0,
#     "anomalous_mag_moment_He3",
#     0.0,
# ]

# !----------------------
# ! Atoms
#
ELEMENT_SYMBOLS = """\
H He Li Be B C N O F Ne Na Mg Al Si P S Cl Ar K Ca Sc
Ti V Cr Mn Fe Co Ni Cu Zn Ga Ge As Se Br Kr Rb Sr Y Zr Nb Mo
Tc Ru Rh Pd Ag Cd In Sn Sb Te I Xe Cs Ba La Ce Pr Nd Pm Sm Eu
Gd Tb Dy Ho Er Tm Yb Lu Hf Ta W Re Os Ir Pt Au Hg Tl Pb Bi Po
At Rn Fr Ra Ac Th Pa U Np Pu Am Cm Bk Cf Es Fm Md No Lr Rf Db
Sg Bh Hs Mt Ds Rg Cn Nh Fl Mc Lv Ts Og
""".strip().split()


# !----------------------
# ! Known Molecules
#
KNOWN_MOLECULES = [
    "CH2",
    "CH3",
    "CH4",
    "CO",
    "CO2",
    "D2",
    "D2O",
    "H2",
    "H2O",
    "N2",
    "HF",
    "OH",
    "O2",
    "NH2",
    "NH3",
    "C2H3",
    "C2H4",
    "C2H5",
]

molecular_mass = [
    14.026579004642441,  # for CH2
    15.034498933118178,  # for CH3
    16.04249886158824,  # for CH4
    28.01009801234052,  # for CO
    44.009496876987235,  # for CO2
    4.028203269749646,  # for D2
    20.02759857879661,  # for D2O
    2.015879856948637,  # for H2
    18.01527872159535,  # for H2O
    28.013398012106343,  # for N2
    20.006341580305058,  # for HF
    17.00733879312103,  # for OH
    31.998797729293425,  # for O2
    16.0228,
    17.030518791476126,  # for NH2, NH3
    27.0457,
    28.0536,
    29.0615,  # For C2H3, C2H4, C2H5
]


@dataclass(frozen=True)
class SubatomicParticle:
    """
    Subatomic particle species.

    Parameters
    ----------
    name : str
        Canonical name of the subatomic particle (e.g., 'electron', 'proton').
    pmd_name : str
        OpenPMD name.
    charge : int
        Particle charge in units of elementary charge.
    mass : float
        Particle mass in eV/c².
    spin : float
    antiparticle : str
    """

    name: str
    pmd_name: str
    charge: int
    mass: float
    spin: float
    antiparticle: str


@dataclass(frozen=True)
class Atom:
    """
    Atomic species.

    Parameters
    ----------
    element : str
        Chemical symbol (e.g., 'C', 'He', 'Au').
    charge : int
        Ionic charge in units of elementary charge.
    nucleon_number : int or None
        Number of nucleons (mass number). If None, uses average isotopic mass.
    is_anti : bool
        Whether this is an anti-atom, default False.
    """

    element: str
    charge: int
    nucleon_number: int | None = None
    is_anti: bool = False


@dataclass(frozen=True)
class Molecule:
    """
    Molecular species.

    Parameters
    ----------
    formula : str
        Molecular formula (e.g., 'H2O', 'NH3').
    charge : int
        Molecular charge in units of elementary charge.
    mass_amu : float or None
        Specified mass in atomic mass units. If None, uses default molecular mass.
    is_known : bool
        Whether this is a known molecule with a predefined formula.
    """

    formula: str
    charge: int
    mass_amu: float | None = None
    is_known: bool = True


@dataclass(frozen=True)
class GenericParticle:
    """
    Generic particle with only mass and charge specified.

    Parameters
    ----------
    mass_amu : float
        Particle mass in atomic mass units.
    charge : int
        Particle charge in units of elementary charge.
    """

    mass_amu: float
    charge: int


# Type alias for any particle species
ParticleSpecies = SubatomicParticle | Atom | Molecule | GenericParticle

# Subatomic particle definitions

SUBATOMIC_PARTICLES = {
    "Anti_Helion": SubatomicParticle("Anti_Helion", "Garbage!", -2, m_helion, 0.0, "Helion"),
    "Anti_Ref_Particle": SubatomicParticle(
        "Anti_Ref_Particle", "Garbage!", 0, 0.0, -987654.3, "Ref_Particle"
    ),
    "Anti_Neutron": SubatomicParticle("Anti_Neutron", "anti-neutron", 0, m_neutron, 0.5, "Neutron"),
    "Anti_Deuteron": SubatomicParticle(
        "Anti_Deuteron", "anti-deuteron", -1, m_deuteron, 1.0, "Deuteron"
    ),
    "Pion-": SubatomicParticle("Pion-", "pion-", -1, m_pion_charged, 0.0, "Pion+"),
    "Muon": SubatomicParticle("Muon", "muon", -1, m_muon, 0.5, "Antimuon"),
    "Antiproton": SubatomicParticle("Antiproton", "anti-proton", -1, m_proton, 0.5, "Proton"),
    "Electron": SubatomicParticle("Electron", "electron", -1, m_electron, 0.5, "Positron"),
    "Photon": SubatomicParticle("Photon", "photon", 0, 0.0, 0.0, "Photon"),
    "Positron": SubatomicParticle("Positron", "positron", 1, m_electron, 0.5, "Electron"),
    "Proton": SubatomicParticle("Proton", "proton", 1, m_proton, 0.5, "Antiproton"),
    "Antimuon": SubatomicParticle("Antimuon", "anti-muon", 1, m_muon, 0.5, "Muon"),
    "Pion+": SubatomicParticle("Pion+", "pion+", 1, m_pion_charged, 0.0, "Pion-"),
    "Deuteron": SubatomicParticle("Deuteron", "deuteron", 1, m_deuteron, 1.0, "Anti_Deuteron"),
    "Neutron": SubatomicParticle("Neutron", "neutron", 0, m_neutron, 0.5, "Anti_Neutron"),
    "Ref_Particle": SubatomicParticle(
        "Ref_Particle", "Garbage!", 0, 0.0, -987654.3, "Anti_Ref_Particle"
    ),
    "Helion": SubatomicParticle("Helion", "Garbage!", 2, m_helion, 0.0, "Anti_Helion"),
    "Pion0": SubatomicParticle("Pion0", "pion0", 0, m_pion_0, 0.0, "Pion0"),
}


LOWER_SUBATOMIC_PARTICLES = {key.lower(): value for key, value in SUBATOMIC_PARTICLES.items()}


def parse_charge(charge_str: str) -> int:
    """
    Parse charge specification from string.

    Parameters
    ----------
    charge_str : str
        Charge specification (e.g., '+', '++', '-3', '+2').

    Returns
    -------
    int
    """
    if not charge_str:
        return 0

    if all(c == "+" for c in charge_str):
        charge = len(charge_str)
    elif all(c == "-" for c in charge_str):
        charge = -len(charge_str)
    elif charge_str[0] in "+-":
        # +N or -N format
        try:
            charge = int(charge_str)
        except ValueError:
            raise InvalidChargeError(f"Invalid charge specification: {charge_str}")
    else:
        raise InvalidChargeError(f"Invalid charge specification: {charge_str}")

    if not -127 <= charge <= 127:
        raise ParticleSpeciesError(f"Charge {charge} out of valid range [-127, 127]")

    return charge


def _parse_mass(mass_str: str) -> float:
    """
    Parse mass specification from @M format.

    Parameters
    ----------
    mass_str : str
        Mass specification after '@M' (e.g., '12.01', '37.5').

    Returns
    -------
    float
        Mass in atomic mass units.
    """
    try:
        mass = float(mass_str)
        if mass < 0:
            raise ParticleSpeciesError(f"Mass cannot be negative: {mass}")
        return mass
    except ValueError:
        raise ParticleSpeciesError(f"Invalid mass specification: {mass_str}")


def parse_species(name: str) -> ParticleSpecies:
    """
    Parse a particle species name into its structured representation.

    Parameters
    ----------
    name : str
        Particle species name following Bmad conventions.

    Returns
    -------
    ParticleSpecies
        Parsed particle species (SubatomicParticle, Atom, Molecule, or GenericParticle).

    Raises
    ------
    ParticleSpeciesError
        If the species name cannot be parsed or is invalid.

    Examples
    --------
    >>> parse_species('electron')
    SubatomicParticle(name='electron', species_id=-1, charge=-1, mass=0.5109989461)

    >>> parse_species('#12C+3')
    Atom(element='C', charge=3, nucleon_number=12, is_anti=False)

    >>> parse_species('H2O+')
    Molecule(formula='H2O', charge=1, mass_amu=None, is_known=True)

    >>> parse_species('@M37.54++')
    GenericParticle(mass_amu=37.54, charge=2)
    """
    if not name or not name.strip():
        raise ParticleSpeciesError("Empty species name")

    original_name = name
    name = name.strip()

    # Handle special cases
    if name.upper() == "REF_SPECIES":
        return LOWER_SUBATOMIC_PARTICLES["ref_particle"]

    # Check for subatomic particle (case insensitive)
    subatomic_key = name.lower()

    subatomic_key = {
        "pion_plus": "pion+",
        "pion_0": "pion0",
        "pion_minus": "pion-",
    }.get(subatomic_key, subatomic_key)

    if subatomic_key in LOWER_SUBATOMIC_PARTICLES:
        return LOWER_SUBATOMIC_PARTICLES[subatomic_key]

    # Parse components: mass (@M), charge (+/-), isotope (#), anti prefix
    remaining = name

    # Extract mass specification (@M)
    mass_amu: float | None = None
    mass_match = re.search(r"@M([\d.]+)", remaining)
    if mass_match:
        mass_amu = _parse_mass(mass_match.group(1))
        remaining = remaining[: mass_match.start()] + remaining[mass_match.end() :]

    # Extract charge specification
    charge = 0
    # Find rightmost + or - that starts a charge specification
    charge_match = re.search(r"([+-](?:\d+|[+-]*))$", remaining)
    if charge_match:
        charge_str = charge_match.group(1)
        # Only parse if there's content after the first +/-
        if len(charge_str) > 0:
            charge = parse_charge(charge_str)
            remaining = remaining[: charge_match.start()]

    # Check for anti prefix
    is_anti = False
    if remaining.startswith("anti"):
        is_anti = True
        remaining = remaining[4:]

    # Extract isotope number (#NNN)
    nucleon_number: int | None = None
    if remaining.startswith("#"):
        iso_match = re.match(r"#(\d*)", remaining)
        if iso_match:
            iso_str = iso_match.group(1)
            nucleon_number = int(iso_str) if iso_str else None
            remaining = remaining[iso_match.end() :]

    # Now determine what type of particle we have
    if not remaining:
        # Only mass and charge specified -> GenericParticle
        if mass_amu is not None:
            return GenericParticle(mass_amu=mass_amu, charge=charge)
        else:
            raise ParticleSpeciesError(f"Cannot determine particle type from: {original_name}")

    # Check if it's a known molecule
    if remaining in KNOWN_MOLECULES:
        if nucleon_number is not None:
            raise ParticleSpeciesError(f"Isotope number not allowed for molecules: {original_name}")
        if is_anti:
            raise ParticleSpeciesError(f"Anti prefix not allowed for molecules: {original_name}")
        return Molecule(formula=remaining, charge=charge, mass_amu=mass_amu, is_known=True)

    # Check if it's an atom
    if remaining in ELEMENT_SYMBOLS or remaining in {"Uut", "Uup", "Uus", "Uuo"}:
        if mass_amu is not None:
            raise ParticleSpeciesError(f"Mass specification not allowed for atoms: {original_name}")
        return Atom(
            element=remaining,
            charge=charge,
            nucleon_number=nucleon_number,
            is_anti=is_anti,
        )

    # If we have a mass specification and remaining text, it's an error
    if mass_amu is not None and remaining:
        raise ParticleSpeciesError(f"Cannot parse species name: {original_name}")

    # Must be an unknown molecule with only mass
    if mass_amu is not None:
        return GenericParticle(mass_amu=mass_amu, charge=charge)

    raise ParticleSpeciesError(f"Cannot parse species name: {original_name}")


def species_to_string(species: ParticleSpecies) -> str:
    """
    Convert a particle species object back to its string representation.

    Parameters
    ----------
    species : ParticleSpecies
        Particle species object.

    Returns
    -------
    str

    Examples
    --------
    >>> species_to_string(SubatomicParticle('electron', -1, -1, 0.511))
    'electron'

    >>> species_to_string(Atom('C', 3, 12, False))
    '#12C+3'

    >>> species_to_string(Molecule('H2O', 1, None, True))
    'H2O+'
    """
    match species:
        case SubatomicParticle(name=name):
            return name

        case Atom(element=element, charge=charge, nucleon_number=n_nuc, is_anti=anti):
            result = ""
            if anti:
                result += "anti"
            if n_nuc is not None:
                result += f"#{n_nuc}"
            result += element
            result += _format_charge(charge)
            return result

        case Molecule(formula=formula, charge=charge, mass_amu=mass, is_known=_):
            result = formula
            if mass is not None:
                result += f"@M{mass:g}"
            result += _format_charge(charge)
            return result

        case GenericParticle(mass_amu=mass, charge=charge):
            result = f"@M{mass:g}"
            result += _format_charge(charge)
            return result


def _format_charge(charge: int) -> str:
    """
    Format charge as string suffix.

    Parameters
    ----------
    charge : int

    Returns
    -------
    str
        Formatted charge string (e.g., '+', '++', '-3', '+2', '').
    """
    if charge == 0:
        return ""
    if charge == 1:
        return "+"
    if charge == -1:
        return "-"
    if charge == 2:
        return "++"
    if charge == -2:
        return "--"
    if charge > 0:
        return f"+{charge}"
    return str(charge)



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class IntrinsicFunction:
    name: str
    arguments: list[str]
    min_args: int
    max_args: int
    description: str


INTRINSIC_FUNCTIONS = {
    "sqrt": IntrinsicFunction(
        name="sqrt",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Square Root",
    ),
    "log": IntrinsicFunction(
        name="log",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Logarithm",
    ),
    "exp": IntrinsicFunction(
        name="exp",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Exponential",
    ),
    "sin": IntrinsicFunction(
        name="sin",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Sine",
    ),
    "cos": IntrinsicFunction(
        name="cos",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Cosine",
    ),
    "tan": IntrinsicFunction(
        name="tan",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Tangent",
    ),
    "cot": IntrinsicFunction(
        name="cot",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Cotangent",
    ),
    "sinc": IntrinsicFunction(
        name="sinc",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Sin(x)/x Function",
    ),
    "asin": IntrinsicFunction(
        name="asin",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Arc sine",
    ),
    "acos": IntrinsicFunction(
        name="acos",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Arc cosine",
    ),
    "atan": IntrinsicFunction(
        name="atan",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Arc tangent",
    ),
    "atan2": IntrinsicFunction(
        name="atan2",
        arguments=["y", "x"],
        min_args=2,
        max_args=2,
        description="Arc tangent of y/x",
    ),
    "sinh": IntrinsicFunction(
        name="sinh",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic sine",
    ),
    "cosh": IntrinsicFunction(
        name="cosh",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic cosine",
    ),
    "tanh": IntrinsicFunction(
        name="tanh",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic tangent",
    ),
    "coth": IntrinsicFunction(
        name="coth",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic cotangent",
    ),
    "asinh": IntrinsicFunction(
        name="asinh",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic arc sine",
    ),
    "acosh": IntrinsicFunction(
        name="acosh",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic arc cosine",
    ),
    "atanh": IntrinsicFunction(
        name="atanh",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic arc tangent",
    ),
    "acoth": IntrinsicFunction(
        name="acoth",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Hyperbolic arc cotangent",
    ),
    "abs": IntrinsicFunction(
        name="abs",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Absolute Value",
    ),
    "factorial": IntrinsicFunction(
        name="factorial",
        arguments=["n"],
        min_args=1,
        max_args=1,
        description="Factorial",
    ),
    "ran": IntrinsicFunction(
        name="ran",
        arguments=[],
        min_args=0,
        max_args=0,
        description="Random number between 0 and 1",
    ),
    "ran_gauss": IntrinsicFunction(
        name="ran_gauss",
        arguments=["sig_cut"],
        min_args=0,
        max_args=1,
        description="Gaussian distributed random number",
    ),
    "int": IntrinsicFunction(
        name="int",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Nearest integer with magnitude less than x",
    ),
    "nint": IntrinsicFunction(
        name="nint",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Nearest integer to x",
    ),
    "sign": IntrinsicFunction(
        name="sign",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="1 if x positive, -1 if negative, 0 if zero",
    ),
    "floor": IntrinsicFunction(
        name="floor",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Nearest integer less than x",
    ),
    "ceiling": IntrinsicFunction(
        name="ceiling",
        arguments=["x"],
        min_args=1,
        max_args=1,
        description="Nearest integer greater than x",
    ),
    "modulo": IntrinsicFunction(
        name="modulo",
        arguments=["a", "p"],
        min_args=2,
        max_args=2,
        description="a - floor(a/p) * p. Will be in range [0, p]",
    ),
    "mass_of": IntrinsicFunction(
        name="mass_of",
        arguments=["A"],
        min_args=1,
        max_args=1,
        description="Mass of particle A",
    ),
    "charge_of": IntrinsicFunction(
        name="charge_of",
        arguments=["A"],
        min_args=1,
        max_args=1,
        description="Charge, in units of the elementary charge, of particle A",
    ),
    "anomalous_moment_of": IntrinsicFunction(
        name="anomalous_moment_of",
        arguments=["A"],
        min_args=1,
        max_args=1,
        description="Anomalous magnetic moment of particle A",
    ),
    "species": IntrinsicFunction(
        name="species",
        arguments=["A"],
        min_args=1,
        max_args=1,
        description="Species ID of A",
    ),
}

RESERVED_NAMES = {
    "beam",
    "beginning",
    "calc_reference_orbit",
    "call",
    "combine_consecutive_elements",
    "debug_marker",
    "end_file",
    "expand_lattice",
    "merge_elements",
    "no_digested",
    "no_superimpose",
    "parameter",
    "parser_debug",
    "particle_start",
    "print",
    "redef",
    "remove_elements",
    "return",
    "root",
    "slice_lattice",
    "start_branch_at",
    "superimpose",
    "title",
    "use",
    "use_local_lat_file",
    "write_digested",
}

BUILTIN_CONSTANTS = {
    "twopi",
    "fourpi",
    "pi",
    "e",
    "e_log",
    "sqrt_2",
    "degrad",
    "degrees",
    "raddeg",
    "m_electron",
    "m_muon",
    "m_pion_0",
    "m_pion_charged",
    "m_proton",
    "m_deuteron",
    "m_neutron",
    "c_light",
    "r_e",
    "r_p",
    "e_charge",
    "h_planck",
    "h_bar_planck",
    "fine_struct_const",
    "anom_moment_electron",
    "anom_moment_proton",
    "anom_moment_neutron",
    "anom_moment_muon",
    "anom_moment_deuteron",
    "anom_moment_he3",
}

ALL_BUILTIN = frozenset(BUILTIN_CONSTANTS | RESERVED_NAMES | set(INTRINSIC_FUNCTIONS))



import inspect

import pytest
import rich

from ..const import (
    AMPERSAND,
    COLON,
    COMMA,
    EQUALS,
    LBRACE,
    LBRACK,
    LPAREN,
    OPEN_TO_CLOSE,
    PLUS,
    RBRACE,
    RBRACK,
    RPAREN,
    SLASH,
    SPACE,
    STAR,
)
from ..exceptions import ExtraCloseDelimiter, MissingCloseDelimiter, UnterminatedString
from ..output import format_statements
from ..parser import (
    Assignment,
    Attribute,
    Block,
    CallName,
    Comments,
    Constant,
    Delimiter,
    Element,
    Empty,
    Line,
    NonstandardParameter,
    Parameter,
    Seq,
    Simple,
    Token,
    UnexpectedAssignment,
    parse,
)
from ..tokenizer import tokenize
from ..types import FormatOptions

T = Token


def parse_verbose(code: str):
    caller = inspect.stack()[1].function
    prefix = f"{caller}:"

    print(prefix)
    print(f"```\n{code}\n```")

    rich.print("\nTokenized:")
    tokens = tokenize(code)
    rich.print("-" * 30)
    rich.print(tokens)
    rich.print("-" * 30)

    res = parse(code)
    rich.print("\nResult:")
    rich.print("-" * 30)
    rich.print(res)
    rich.print("-" * 30)

    return res


def test_from_items_tokens():
    items = [Token("a"), Token("b"), Token("c")]
    res = Seq.from_items(list(items))
    assert res == Seq(
        opener=None, closer=None, items=[Token("a"), Token("b"), Token("c")], delimiter=SPACE
    )


def test_from_items_block():
    opener = LBRACE
    closer = RBRACE
    items = [Block(opener=opener, closer=closer, items=[Token("a"), Token("b"), Token("c")])]
    res = Seq.from_items(list(items))
    assert res == Seq(
        opener=opener,
        closer=closer,
        items=[Seq(items=[Token("a"), Token("b"), Token("c")], delimiter=SPACE)],
        delimiter=COMMA,
    )


def test_from_items_delimited_block():
    opener = LBRACE
    closer = RBRACE
    items = [
        Block(
            opener=opener, closer=closer, items=[Token("a"), COMMA, Token("b"), COMMA, Token("c")]
        )
    ]
    res = Seq.from_items(list(items))
    assert res == Seq(
        opener=opener, closer=closer, items=[Token("a"), Token("b"), Token("c")], delimiter=COMMA
    )


def test_from_items_delimited_nested_block():
    opener = LBRACE
    closer = RBRACE

    inner_tokens = [Token("a"), Token("b"), Token("c")]
    blk = Block(opener=opener, closer=closer, items=list(inner_tokens))
    items = [Block(opener=opener, closer=closer, items=[Token("a"), COMMA, Token("b"), COMMA, blk])]
    res = Seq.from_items(list(items))
    assert res == Seq(
        opener=opener,
        closer=closer,
        items=list(
            [
                Token("a"),
                Token("b"),
                Seq(
                    opener=opener,
                    closer=closer,
                    items=[Seq(items=list(inner_tokens), delimiter=SPACE)],
                    delimiter=COMMA,
                ),
            ]
        ),
        delimiter=COMMA,
    )


def test_from_items_named_token():
    items = [Token("a"), EQUALS, Token("c")]
    res = Seq.from_items(items)
    assert res == Attribute(name=Token("a"), value=Token("c"))


def test_from_items_named_sequence_curly():
    opener = LBRACE
    closer = RBRACE

    inner_tokens = [Token("a"), Token("b"), Token("c")]
    blk = Block(opener=opener, closer=closer, items=list(inner_tokens))
    items = [Token("a"), EQUALS, blk]
    res = Seq.from_items(items)
    assert res == Attribute(
        name=Token("a"),
        value=Seq(
            opener=opener,
            closer=closer,
            items=[Seq(items=list(inner_tokens), delimiter=SPACE)],
            delimiter=COMMA,
        ),
    )


def test_from_items_named_sequence_curly_comma_delimited():
    opener = LBRACE
    closer = RBRACE
    blk = Block(
        opener=opener, closer=closer, items=[Token("a"), COMMA, Token("b"), COMMA, Token("c")]
    )
    items = [Token("a"), EQUALS, blk]
    res = Seq.from_items(items)
    assert res == Attribute(
        name=Token("a"),
        value=Seq(
            opener=opener,
            closer=closer,
            items=[Token("a"), Token("b"), Token("c")],
            delimiter=COMMA,
        ),
    )


def make_seq(opener: str, items, delimiter: str | None = COMMA) -> Seq:
    return Seq(
        opener=Delimiter(opener),
        closer=Delimiter(OPEN_TO_CLOSE[opener]),
        items=items,
        delimiter=Delimiter(delimiter) if delimiter else None,
    )


def make_block(opener: str, items) -> Seq:
    if not opener:
        return Seq(items=items)

    return Seq(opener=Delimiter(opener), closer=Delimiter(OPEN_TO_CLOSE[opener]), items=items)


def test_extra_close_delimiter():
    line = "parameter[foo]        ]"
    with pytest.raises(ExtraCloseDelimiter) as raised:
        assert parse_verbose(line)
    print(raised)
    delim = raised.value.delim
    assert delim == "]"
    assert delim.loc.column == len(line) - 1
    assert delim.loc.end_column == len(line)
    assert line[delim.loc.column : delim.loc.end_column] == delim
    assert delim.loc.line == 0
    assert delim.loc.end_line == 0


def test_missing_close_delimiter():
    line = "parameter[foo"
    with pytest.raises(MissingCloseDelimiter) as raised:
        assert parse_verbose(line)
    print(raised)

    (delim,) = raised.value.delims
    assert delim == "["
    assert delim.loc.column == line.index("[")
    assert delim.loc.end_column == line.index("[") + 1
    assert line[delim.loc.column : delim.loc.end_column] == delim
    assert delim.loc.line == 0
    assert delim.loc.end_line == 0


def roundtrip_code(code: str) -> None:
    options = FormatOptions()
    # code1 -> tokens1 -> roundtrip_code -> roundtrip_tokens
    #  check tokens1 == tokens2
    tokens1 = parse_verbose(code)

    roundtrip_code = format_statements(tokens1, options)

    rich.print("Back to string:")
    rich.print("=" * 30)
    print(roundtrip_code)
    rich.print("=" * 30)

    roundtrip_tokens = parse(roundtrip_code)
    assert tokens1 == roundtrip_tokens


code_tests = pytest.mark.parametrize(
    ("code", "expected"),
    [
        pytest.param(
            """
    global_rf_frequency = 1.3e9
    """,
            [
                Constant(
                    name=T("global_rf_frequency"),
                    value=T("1.3e9"),
                )
            ],
            id="constant_without_comments",
        ),
        pytest.param(
            """
    global_rf_frequency &
    = &
    1.3e9
    """,
            [
                Constant(
                    name=T("global_rf_frequency"),
                    value=T("1.3e9"),
                )
            ],
            id="excessive_continuation",
        ),
        pytest.param(
            """
    global_rf_frequency := 1.3e9
    """,
            [
                Constant(
                    name=T("global_rf_frequency"),
                    value=T("1.3e9"),
                )
            ],
            id="constant_using_walrus_operator",
        ),
        pytest.param(
            "my_const = sqrt(10.3 + 1) * pi^3",
            [
                Constant(
                    name=T("my_const"),
                    value=Seq(
                        items=[
                            T("sqrt"),
                            make_seq(
                                "(",
                                items=[
                                    Seq(
                                        items=[
                                            T("10.3"),
                                            T("+"),
                                            T("1"),
                                        ],
                                        delimiter=SPACE,
                                    ),
                                ],
                                delimiter=COMMA,
                            ),
                            T("*"),
                            T("pi"),
                            T("^"),
                            T("3"),
                        ],
                        delimiter=SPACE,
                    ),
                )
            ],
            id="constant_with_function_and_power",
        ),
        pytest.param(
            "abc := my_const * 23",
            [
                Constant(
                    name=T("abc"),
                    value=Seq(items=[T("my_const"), T("*"), T("23")], delimiter=SPACE),
                )
            ],
            id="constant_with_colon_equals",
        ),
        pytest.param(
            """
    !pre
    global_rf_frequency = 1.3e9  !inline
    """,
            [
                Constant(
                    comments=Comments(pre=[T("pre")], inline=T("inline")),
                    name=T("global_rf_frequency"),
                    value=T("1.3e9"),
                )
            ],
            id="constant_with_comments",
        ),
        pytest.param(
            """
    !pre
    global_rf_frequency = 1.3e9  !inline
    !post1
    !post2
    !post3
    """,
            [
                Constant(
                    comments=Comments(pre=[T("pre")], inline=T("inline")),
                    name=T("global_rf_frequency"),
                    value=T("1.3e9"),
                ),
                Empty(
                    comments=Comments(
                        pre=[
                            T("post1"),
                            T("post2"),
                            T("post3"),
                        ]
                    ),
                ),
            ],
            id="post_comments",
        ),
        pytest.param(
            """
    !pre
    redef: global_rf_frequency = 1.3e9  !inline
    """,
            [
                Constant(
                    comments=Comments(pre=[T("pre")], inline=T("inline")),
                    name=T("global_rf_frequency"),
                    value=T("1.3e9"),
                    redef=True,
                )
            ],
            id="constant_redef",
        ),
        pytest.param(
            """
    !pre
    my_const = b01w[rho] / b01w[L] ! Use of attribute values in an expression.
    """,
            [
                Constant(
                    comments=Comments(
                        pre=[T("pre")],
                        inline=T(T(" Use of attribute values in an expression.")),
                    ),
                    name=T("my_const"),
                    value=Seq(
                        items=[
                            T("b01w"),
                            make_block(opener="[", items=[T("rho")]),
                            Token("/"),
                            T("b01w"),
                            make_block(opener="[", items=[T("L")]),
                        ],
                    ),
                )
            ],
            id="constant_with_attribute_value",
        ),
        pytest.param(
            "x_ray: line[multipass] = (el1, el2, el3)",
            [
                Line(
                    name=T("x_ray"),
                    elements=make_seq(
                        opener=LPAREN,
                        items=[
                            T("el1"),
                            T("el2"),
                            T("el3"),
                        ],
                        delimiter=",",
                    ),
                    multipass=True,
                )
            ],
            id="line_element_multipass",
        ),
        pytest.param(
            "x_ray: line = (el1, el2, el3)",
            [
                Line(
                    name=T("x_ray"),
                    elements=make_seq(
                        opener=LPAREN,
                        items=[
                            T("el1"),
                            T("el2"),
                            T("el3"),
                        ],
                        delimiter=",",
                    ),
                )
            ],
            id="line_element",
        ),
        pytest.param(
            "arg_ln(a, b): line = (ele1, a, ele2, b)",
            [
                Line(
                    name=CallName(
                        name=T("arg_ln"),
                        args=make_seq(opener="(", items=[T("a"), T("b")], delimiter=","),
                    ),
                    elements=make_seq(
                        opener="(",
                        items=[
                            T("ele1"),
                            T("a"),
                            T("ele2"),
                            T("b"),
                        ],
                        delimiter=",",
                    ),
                )
            ],
            id="line_element_with_arguments",
        ),
        pytest.param(
            """
    betweenhalf00(sfw,sdw,ofw,odw) : line=(  dbt1, qbt1,qbt1, dbt2,  qbt2,qbt2, dbt3,  &
            ofw, sfw, dbt4, qbt3,qbt3, dbt5, qbt4,qbt4, dbt6, qbt5,qbt5, dbt7, odw, sdw ) ;
    """,
            [
                Line(
                    name=CallName(
                        name=T("betweenhalf00"),
                        args=make_seq(
                            opener="(",
                            items=[
                                T("sfw"),
                                T("sdw"),
                                T("ofw"),
                                T("odw"),
                            ],
                            delimiter=",",
                        ),
                    ),
                    elements=make_seq(
                        opener=LPAREN,
                        items=[
                            T(ele)
                            for ele in "dbt1,qbt1,qbt1,dbt2,qbt2,qbt2,dbt3,ofw,sfw,dbt4,qbt3,qbt3,dbt5,qbt4,qbt4,dbt6,qbt5,qbt5,dbt7,odw,sdw".split(
                                ","
                            )
                        ],
                        delimiter=",",
                    ),
                )
            ],
            id="line_element_with_arguments_wild",
        ),
        pytest.param(
            """
    ! pre
    cst1: crystal, b_param = -0.6, crystal_type = "Si(620)", use_reflectivity_table = T,
            reflectivity_table = {polarization = sigma, call::reflect_pi.bmad},
            reflectivity_table = {polarization = pi, call::reflect_pi.bmad}
    """,
            [
                Element(
                    comments=Comments(pre=[T(" pre")]),
                    name=T("cst1"),
                    keyword=T("crystal"),
                    attributes=[
                        Attribute(name=T("b_param"), value=T("-0.6")),
                        Attribute(name=T("crystal_type"), value=T('"Si(620)"')),
                        Attribute(
                            name=T("use_reflectivity_table"),
                            value=T("T"),
                        ),
                        Attribute(
                            name=T("reflectivity_table"),
                            value=make_seq(
                                opener=LBRACE,
                                items=[
                                    Attribute(
                                        name=T("polarization"),
                                        value=T("sigma"),
                                    ),
                                    Seq(
                                        delimiter=SPACE,
                                        items=[T("call"), Delimiter("::"), T("reflect_pi.bmad")],
                                    ),
                                ],
                            ),
                        ),
                        Attribute(
                            name=T("reflectivity_table"),
                            value=make_seq(
                                opener=LBRACE,
                                items=[
                                    Attribute(
                                        name=T("polarization"),
                                        value=T("pi"),
                                    ),
                                    Seq(
                                        delimiter=SPACE,
                                        items=[T("call"), Delimiter("::"), T("reflect_pi.bmad")],
                                    ),
                                ],
                            ),
                        ),
                    ],
                )
            ],
            id="crystal_element_with_reflectivity_tables",
        ),
        pytest.param(
            """
cst3: crystal, b_param =  0.6, crystal_type = "Ge(111)",
  h_misalign = {
    ix_bounds = (-1, 1),
    iy_bounds = (-2, 0),
    r0 = (0.01, 0.02),
    dr = (0.02, 0.03),
    pt(-1, -2) = (0.0010, 0.0020, 0.0030, 0.0040),
    pt(-1, -1) = (0.0011, 0.0021, 0.0031, 0.0041),
  }
        """,
            [
                Element(
                    name=T("cst3"),
                    keyword=T("crystal"),
                    attributes=[
                        Attribute(name=T("b_param"), value=T("0.6")),
                        Attribute(name=T("crystal_type"), value=T('"Ge(111)"')),
                        Attribute(
                            name=T("h_misalign"),
                            value=make_seq(
                                "{",
                                items=[
                                    Attribute(
                                        name=T("ix_bounds"),
                                        value=make_seq(
                                            opener="(",
                                            items=[T("-1"), T("1")],
                                        ),
                                    ),
                                    Attribute(
                                        name=T("iy_bounds"),
                                        value=make_seq(
                                            opener="(",
                                            items=[T("-2"), T("0")],
                                        ),
                                    ),
                                    Attribute(
                                        name=T("r0"),
                                        value=make_seq(
                                            opener="(",
                                            items=[T("0.01"), T("0.02")],
                                        ),
                                    ),
                                    Attribute(
                                        name=T("dr"),
                                        value=make_seq(
                                            opener="(",
                                            items=[T("0.02"), T("0.03")],
                                        ),
                                    ),
                                    Attribute(
                                        name=CallName(
                                            name=T("pt"),
                                            args=make_seq(opener="(", items=[T("-1"), T("-2")]),
                                        ),
                                        value=make_seq(
                                            opener="(",
                                            items=[
                                                T("0.0010"),
                                                T("0.0020"),
                                                T("0.0030"),
                                                T("0.0040"),
                                            ],
                                        ),
                                    ),
                                    Attribute(
                                        name=CallName(
                                            name=T("pt"),
                                            args=make_seq(opener="(", items=[T("-1"), T("-1")]),
                                        ),
                                        value=make_seq(
                                            opener="(",
                                            items=[
                                                T("0.0011"),
                                                T("0.0021"),
                                                T("0.0031"),
                                                T("0.0041"),
                                            ],
                                        ),
                                    ),
                                ],
                            ),
                        ),
                    ],
                )
            ],
            id="attribute_with_call_args",
        ),
        pytest.param(
            """
        parameter[lattice] = CBETA
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("lattice"),
                    value=T('"CBETA"'),
                ),
            ],
            id="parameter_lattice_unquoted",
        ),
        pytest.param(
            """
        parameter[lattice] = "CBETA"
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("lattice"),
                    value=T('"CBETA"'),
                ),
            ],
            id="parameter_lattice_quoted",
        ),
        pytest.param(
            """
        parameter[default_tracking_species] = He++
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("default_tracking_species"),
                    value=T("He++"),
                ),
            ],
            id="parameter_default_tracking_species",
        ),
        pytest.param(
            """
        parameter[live_branch] = t
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("live_branch"),
                    value=T("True"),
                ),
            ],
            id="parameter_live_branch_bool_true",
        ),
        pytest.param(
            """
        parameter[live_branch] = f
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("live_branch"),
                    value=T("False"),
                ),
            ],
            id="parameter_live_branch_bool_false",
        ),
        pytest.param(
            """
        parameter[machine] = machine-name-with-dashes-because:why-not
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("machine"),
                    value=T('"machine-name-with-dashes-because:why-not"'),
                ),
            ],
            id="parameter_dumb_machine_name",
        ),
        pytest.param(
            """
        parameter[lattice] = lattice-name-with-dashes-because:why-not
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("lattice"),
                    value=T('"lattice-name-with-dashes-because:why-not"'),
                ),
            ],
            id="parameter_dumb_lattice_name",
        ),
        pytest.param(
            """
        parameter[no_end_marker] = f
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("no_end_marker"),
                    value=T("False"),
                ),
            ],
            id="parameter_no_end_marker",
        ),
        pytest.param(
            """
        parameter[p0c] = 1.2
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("p0c"),
                    value=T("1.2"),
                ),
            ],
            id="parameter_p0c",
        ),
        pytest.param(
            """
        parameter[ran_seed] = 1
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("ran_seed"),
                    value=T("1"),
                ),
            ],
            id="parameter_ran_seed",
        ),
        pytest.param(
            """
        parameter[taylor_order] = 1
        """,
            [
                Parameter(
                    target=T("parameter"),
                    name=T("taylor_order"),
                    value=T("1"),
                ),
            ],
            id="parameter_taylor_order",
        ),
        pytest.param(
            """
        beginning[e_tot] = 6e6
        """,
            [
                Parameter(
                    target=T("beginning"),
                    name=T("e_tot"),
                    value=T("6e6"),
                ),
            ],
            id="beginning",
        ),
        pytest.param(
            """
        q2: quadrupole, tracking_method = symp_lie_ptc
        q2[tracking_method] = symp_lie_ptc
        quadrupole::*[tracking_method] = symp_lie_ptc
        """,
            [
                Element(
                    name=T("q2"),
                    keyword=T("quadrupole"),
                    attributes=[Attribute(name=T("tracking_method"), value=T("symp_lie_ptc"))],
                ),
                Parameter(
                    target=T("q2"),
                    name=T("tracking_method"),
                    value=T("symp_lie_ptc"),
                ),
                Parameter(
                    target=Seq(delimiter=SPACE, items=[T("quadrupole"), Delimiter("::"), T("*")]),
                    name=T("tracking_method"),
                    value=T("symp_lie_ptc"),
                ),
            ],
            id="manual_chap6_intro",
        ),
        pytest.param(
            "b01w[roll] = 6.5 ! Set an attribute value.",
            [
                Parameter(
                    comments=Comments(inline=T(T(" Set an attribute value."))),
                    target=T("b01w"),
                    name=T("roll"),
                    value=T("6.5"),
                )
            ],
            id="element_parameter_assignment_roll",
        ),
        pytest.param(
            "b01w[L] = 6.5 ! Change an attribute value.",
            [
                Parameter(
                    comments=Comments(inline=T(T(" Change an attribute value."))),
                    target=T("b01w"),
                    name=T("L"),
                    value=T("6.5"),
                )
            ],
            id="element_parameter_assignment_L",
        ),
        pytest.param(
            "b01w[L] = b01w[rho] / 12 ! OK to reset an attribute value.",
            [
                Parameter(
                    comments=Comments(inline=T(T(" OK to reset an attribute value."))),
                    target=T("b01w"),
                    name=T("L"),
                    value=Seq(
                        items=[
                            T("b01w"),
                            make_seq(
                                opener=LBRACK,
                                items=[T("rho")],
                                delimiter=SPACE,
                            ),
                            T("/"),
                            T("12"),
                        ],
                    ),
                )
            ],
            id="element_parameter_assignment_expression",
        ),
        pytest.param(
            "97[x_offset] = 0.0023 ! Set x_offset attribute of 97th element",
            [
                Parameter(
                    comments=Comments(inline=T(T(" Set x_offset attribute of 97th element"))),
                    target=T("97"),
                    name=T("x_offset"),
                    value=T("0.0023"),
                )
            ],
            id="element_parameter_assignment_numeric_target",
        ),
        pytest.param(
            'b2>>si_cryst##2[tilt] = 0.1 ! Tilt the 2nd instance of "si_cryst" in branch "b2"',
            [
                Parameter(
                    comments=Comments(
                        inline=T(T(' Tilt the 2nd instance of "si_cryst" in branch "b2"'))
                    ),
                    target=T("b2>>si_cryst##2"),
                    name=T("tilt"),
                    value=T("0.1"),
                )
            ],
            id="element_parameter_assignment_branch_instance",
        ),
        pytest.param(
            "5:32[x_limit] = 0.3 ! Sets elements with indexes 5 through 32 in branch 0.",
            [
                Parameter(
                    comments=Comments(
                        inline=T(T(" Sets elements with indexes 5 through 32 in branch 0."))
                    ),
                    target=Seq(items=[T("5"), T(":"), T("32")], delimiter=SPACE),
                    name=T("x_limit"),
                    value=T("0.3"),
                )
            ],
            id="element_parameter_assignment_range",
        ),
        pytest.param(
            "parameter[custom_attribute1] = quadrupole::error_k1",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("custom_attribute1"),
                    value=Seq(
                        delimiter=SPACE, items=[T("quadrupole"), Delimiter("::"), T("error_k1")]
                    ),
                )
            ],
            id="custom_element_attributes_quadrupole_error",
        ),
        pytest.param(
            "parameter[custom_attribute1] = mag_id",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("custom_attribute1"),
                    value=T("mag_id"),
                )
            ],
            id="custom_element_attributes_mag_id",
        ),
        pytest.param(
            "parameter[custom_attribute1] = sextupole::error_k2",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("custom_attribute1"),
                    value=Seq(
                        delimiter=SPACE, items=[T("sextupole"), Delimiter("::"), T("error_k2")]
                    ),
                )
            ],
            id="custom_element_attributes_sextupole_error",
        ),
        pytest.param(
            "parameter[custom_attribute2] = color",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("custom_attribute2"),
                    value=T("color"),
                )
            ],
            id="custom_element_attributes_color",
        ),
        pytest.param(
            "parameter[custom_attribute2] = parameter::quad_mag_moment",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("custom_attribute2"),
                    value=Seq(
                        delimiter=SPACE,
                        items=[T("parameter"), Delimiter("::"), T("quad_mag_moment")],
                    ),
                )
            ],
            id="custom_element_attributes_quad_mag_moment",
        ),
        pytest.param(
            "qq: quadrupole, r_custom(-2,1,5) = 34.5, r_custom(-3) = 77.9",
            [
                Element(
                    name=T("qq"),
                    keyword=T("quadrupole"),
                    attributes=[
                        Attribute(
                            name=CallName(
                                name=T("r_custom"),
                                args=make_seq(opener="(", items=[T("-2"), T("1"), T("5")]),
                            ),
                            value=T("34.5"),
                        ),
                        Attribute(
                            name=CallName(
                                name=T("r_custom"),
                                args=make_seq(opener="(", items=[T("-3")], delimiter=COMMA),
                            ),
                            value=T("77.9"),
                        ),
                    ],
                )
            ],
            id="negative_indices",
        ),
        pytest.param(
            "parameter[particle] = #12C+3",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("particle"),
                    value=T("#12C+3"),
                )
            ],
            id="species_names_triply_charged_carbon",
        ),
        pytest.param(
            "parameter[p0c] = 12 * 500e6",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("p0c"),
                    value=Seq(
                        items=[
                            T("12"),
                            T("*"),
                            T("500e6"),
                        ],
                    ),
                )
            ],
            id="species_names_reference_momentum",
        ),
        pytest.param(
            "parameter[particle] = He--",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("particle"),
                    value=T("He--"),
                )
            ],
            id="species_names_doubly_charged_he",
        ),
        pytest.param(
            "parameter[particle] = C2H3@M28.4+",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("particle"),
                    value=T("C2H3@M28.4+"),
                )
            ],
            id="species_names_singly_charged_c2h3",
        ),
        pytest.param(
            "parameter[particle] = CH2",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("particle"),
                    value=T("CH2"),
                )
            ],
            id="species_names_natural_ch2",
        ),
        pytest.param(
            "parameter[particle] = @M37.54++",
            [
                Parameter(
                    target=T("parameter"),
                    name=T("particle"),
                    value=T("@M37.54++"),
                )
            ],
            id="species_names_doubly_charged_molecule",
        ),
        pytest.param(
            """
! Overlay element definition with internal comments
ov1: overlay = {  ! ele_comment
    q1,  ! list_comment
    q2
}, master_attr = 3.4, flag_attr  ! flag_attr comment
    """,
            [
                Element(
                    comments=Comments(
                        pre=[T(" Overlay element definition with internal comments")],
                        inline=T(" ele_comment"),
                    ),
                    name=T("ov1"),
                    keyword=T("overlay"),
                    ele_list=Seq(
                        opener=LBRACE,
                        closer=RBRACE,
                        items=[
                            T("q1", comments=Comments(inline=T(" list_comment"))),
                            T("q2"),
                        ],
                        delimiter=COMMA,
                    ),
                    attributes=[
                        Attribute(
                            name=T(
                                "master_attr",
                            ),
                            value=T("3.4"),
                        ),
                        Attribute(
                            name=T(
                                "flag_attr",
                                comments=Comments(inline=T(" flag_attr comment")),
                            ),
                            value=None,
                        ),
                    ],
                )
            ],
            id="overlay_parsing",
        ),
        pytest.param(
            """
A.B.Gat02: IN.CM1.Gat00,  !inline
    wall = {
section = { s = 0,  !0-comment
  v(1) = {0, 0, A.B.Gat02[aperture]}},  !aperture-comment
section = { s = A.B.Gat02[L],
  v(1) = {0, 0, A.B.Gat02[aperture]}}}
""",
            [
                Element(
                    comments=Comments(inline=Token("inline")),
                    name=Token("A.B.Gat02"),
                    keyword=Token("IN.CM1.Gat00"),
                    ele_list=None,
                    attributes=[
                        Attribute(
                            name=Token("wall"),
                            value=Seq(
                                opener=LBRACE,
                                closer=RBRACE,
                                items=[
                                    Attribute(
                                        name=Token("section"),
                                        value=Seq(
                                            opener=LBRACE,
                                            closer=RBRACE,
                                            items=[
                                                Attribute(
                                                    name=Token("s"),
                                                    value=Token(
                                                        "0",
                                                        comments=Comments(
                                                            inline=Token("0-comment")
                                                        ),
                                                    ),
                                                ),
                                                Attribute(
                                                    name=CallName(
                                                        name=Token("v"),
                                                        args=Seq(
                                                            opener=LPAREN,
                                                            closer=RPAREN,
                                                            items=[Token("1")],
                                                            delimiter=COMMA,
                                                        ),
                                                    ),
                                                    value=Seq(
                                                        opener=LBRACE,
                                                        closer=RBRACE,
                                                        items=[
                                                            Token("0"),
                                                            Token("0"),
                                                            Seq(
                                                                opener=None,
                                                                closer=None,
                                                                items=[
                                                                    Token("A.B.Gat02"),
                                                                    Seq(
                                                                        opener=LBRACK,
                                                                        closer=RBRACK,
                                                                        items=[
                                                                            Token(
                                                                                "aperture",
                                                                                comments=Comments(
                                                                                    inline=Token(
                                                                                        "aperture-comment"
                                                                                    )
                                                                                ),
                                                                            )
                                                                        ],
                                                                        delimiter=SPACE,
                                                                    ),
                                                                ],
                                                                delimiter=SPACE,
                                                            ),
                                                        ],
                                                        delimiter=COMMA,
                                                    ),
                                                ),
                                            ],
                                            delimiter=COMMA,
                                        ),
                                    ),
                                    Attribute(
                                        name=Token("section"),
                                        value=Seq(
                                            opener=LBRACE,
                                            closer=RBRACE,
                                            items=[
                                                Attribute(
                                                    name=Token("s"),
                                                    value=Seq(
                                                        opener=None,
                                                        closer=None,
                                                        items=[
                                                            Token("A.B.Gat02"),
                                                            Seq(
                                                                opener=LBRACK,
                                                                closer=RBRACK,
                                                                items=[Token("L")],
                                                                delimiter=SPACE,
                                                            ),
                                                        ],
                                                        delimiter=SPACE,
                                                    ),
                                                ),
                                                Attribute(
                                                    name=CallName(
                                                        name=Token("v"),
                                                        args=Seq(
                                                            opener=LPAREN,
                                                            closer=RPAREN,
                                                            items=[Token("1")],
                                                            delimiter=COMMA,
                                                        ),
                                                    ),
                                                    value=Seq(
                                                        opener=LBRACE,
                                                        closer=RBRACE,
                                                        items=[
                                                            Token("0"),
                                                            Token("0"),
                                                            Seq(
                                                                opener=None,
                                                                closer=None,
                                                                items=[
                                                                    Token("A.B.Gat02"),
                                                                    Seq(
                                                                        opener=LBRACK,
                                                                        closer=RBRACK,
                                                                        items=[Token("aperture")],
                                                                        delimiter=SPACE,
                                                                    ),
                                                                ],
                                                                delimiter=SPACE,
                                                            ),
                                                        ],
                                                        delimiter=Delimiter(","),
                                                    ),
                                                ),
                                            ],
                                            delimiter=Delimiter(","),
                                        ),
                                    ),
                                ],
                                delimiter=Delimiter(","),
                            ),
                        )
                    ],
                )
            ],
            id="wall_parsing",
        ),
        pytest.param(
            """
QTov1 : overlay = {QT1[k1]:k1*dk , QT4[k1]:k1*dk , QT13[k1]:k1*dk, QT16[k1]:k1*dk}, var = {k1,dk}, k1= -0.553299, dk=1.0
""",
            [
                Element(
                    name=T("QTov1"),
                    keyword=T("overlay"),
                    ele_list=Seq(
                        opener=LBRACE,
                        closer=RBRACE,
                        items=[
                            Seq(
                                items=[
                                    T("QT1"),
                                    Seq(
                                        opener=LBRACK,
                                        closer=RBRACK,
                                        items=[T("k1")],
                                    ),
                                    COLON,
                                    T("k1"),
                                    STAR,
                                    T("dk"),
                                ],
                            ),
                            Seq(
                                items=[
                                    T("QT4"),
                                    Seq(
                                        opener=LBRACK,
                                        closer=RBRACK,
                                        items=[T("k1")],
                                    ),
                                    COLON,
                                    T("k1"),
                                    STAR,
                                    T("dk"),
                                ],
                            ),
                            Seq(
                                items=[
                                    T("QT13"),
                                    Seq(
                                        opener=LBRACK,
                                        closer=RBRACK,
                                        items=[T("k1")],
                                    ),
                                    COLON,
                                    T("k1"),
                                    STAR,
                                    T("dk"),
                                ],
                            ),
                            Seq(
                                items=[
                                    T("QT16"),
                                    Seq(
                                        opener=LBRACK,
                                        closer=RBRACK,
                                        items=[T("k1")],
                                    ),
                                    COLON,
                                    T("k1"),
                                    STAR,
                                    T("dk"),
                                ],
                            ),
                        ],
                        delimiter=COMMA,
                    ),
                    attributes=[
                        Attribute(
                            name=T("var"),
                            value=Seq(
                                opener=LBRACE,
                                closer=RBRACE,
                                items=[T("k1"), T("dk")],
                                delimiter=COMMA,
                            ),
                        ),
                        Attribute(name=T("k1"), value=T("-0.553299")),
                        Attribute(name=T("dk"), value=T("1.0")),
                    ],
                )
            ],
            id="overlay_with_expressions",
        ),
        pytest.param(
            """
HB01: overlay = {B04W /0.5, B05W /0.5}, HKICK, Alias = "HB1", &
        Type = "CSR HBND CUR   1"
""",
            [
                Element(
                    name=T("HB01"),
                    keyword=T("overlay"),
                    ele_list=Seq(
                        opener=LBRACE,
                        closer=RBRACE,
                        items=[
                            Seq(items=[T("B04W"), SLASH, T("0.5")]),
                            Seq(items=[T("B05W"), SLASH, T("0.5")]),
                        ],
                        delimiter=COMMA,
                    ),
                    attributes=[
                        Attribute(name=T("HKICK"), value=None),
                        Attribute(name=T("Alias"), value=T('"HB1"')),
                        Attribute(name=T("Type"), value=T('"CSR HBND CUR   1"')),
                    ],
                )
            ],
            id="overlay_with_division_weight",
        ),
        pytest.param(
            """
O_B7_b12: overlay = {B7, B12, B4:-g*b7[l]/b4[l]}, var = {g}
""",
            [
                Element(
                    name=T("O_B7_b12"),
                    keyword=T("overlay"),
                    ele_list=Seq(
                        opener=LBRACE,
                        closer=RBRACE,
                        delimiter=COMMA,
                        items=[
                            T("B7"),
                            T("B12"),
                            Seq(
                                items=[
                                    T("B4"),
                                    COLON,
                                    Delimiter("-"),
                                    T("g"),
                                    STAR,
                                    T("b7"),
                                    make_seq(opener=LBRACK, items=[T("l")], delimiter=SPACE),
                                    SLASH,
                                    T("b4"),
                                    make_seq(opener=LBRACK, items=[T("l")], delimiter=SPACE),
                                ],
                            ),
                        ],
                    ),
                    attributes=[
                        Attribute(
                            name=T("var"),
                            value=Seq(
                                opener=LBRACE,
                                closer=RBRACE,
                                items=[T("g")],
                                delimiter=COMMA,
                            ),
                        )
                    ],
                )
            ],
            id="overlay_negative_expression",
        ),
        pytest.param(
            """
delta1: group = {len1[L]:0.23061/360*deg}, var={deg}
""",
            [
                Element(
                    name=T("delta1"),
                    keyword=T("group"),
                    ele_list=Seq(
                        opener=LBRACE,
                        closer=RBRACE,
                        delimiter=COMMA,
                        items=[
                            Seq(
                                delimiter=SPACE,
                                items=[
                                    T("len1"),
                                    make_seq(opener=LBRACK, items=[T("L")], delimiter=SPACE),
                                    COLON,
                                    T("0.23061"),
                                    SLASH,
                                    T("360"),
                                    STAR,
                                    T("deg"),
                                ],
                            )
                        ],
                    ),
                    attributes=[
                        Attribute(
                            name=T("var"),
                            value=Seq(
                                opener=LBRACE,
                                closer=RBRACE,
                                items=[T("deg")],
                                delimiter=COMMA,
                            ),
                        )
                    ],
                )
            ],
            id="group_with_arithmetic",
        ),
        pytest.param(
            """
fig8 : line  = (ir1, L4, ir2, ir1, p1, --L3, p2, ir2);
""",
            [
                Line(
                    name=T("fig8"),
                    elements=Seq(
                        opener=LPAREN,
                        closer=RPAREN,
                        delimiter=COMMA,
                        items=[
                            T("ir1"),
                            T("L4"),
                            T("ir2"),
                            T("ir1"),
                            T("p1"),
                            Seq(
                                items=[
                                    Delimiter("--"),
                                    T("L3"),
                                ],
                                delimiter=SPACE,
                            ),
                            T("p2"),
                            T("ir2"),
                        ],
                    ),
                    multipass=False,
                )
            ],
            id="reversed_line",
        ),
        pytest.param(
            """
cell3 : LINE=  (d1, qd, d2, bb, d2, qf, d1)
L3 : LINE = (bb, d2, qf, d1, 8*cell3, d1, qd, d2, bb)
""",
            [
                Line(
                    name=T("cell3"),
                    elements=Seq(
                        opener=LPAREN,
                        closer=RPAREN,
                        items=[
                            T("d1"),
                            T("qd"),
                            T("d2"),
                            T("bb"),
                            T("d2"),
                            T("qf"),
                            T("d1"),
                        ],
                        delimiter=COMMA,
                    ),
                    multipass=False,
                ),
                Line(
                    name=T("L3"),
                    elements=Seq(
                        opener=LPAREN,
                        closer=RPAREN,
                        items=[
                            T("bb"),
                            T("d2"),
                            T("qf"),
                            T("d1"),
                            Seq(
                                items=[
                                    T("8"),
                                    STAR,
                                    T("cell3"),
                                ],
                                delimiter=SPACE,
                            ),
                            T("d1"),
                            T("qd"),
                            T("d2"),
                            T("bb"),
                        ],
                        delimiter=COMMA,
                    ),
                    multipass=False,
                ),
            ],
            id="line_with_repetition",
        ),
        pytest.param(
            """
b  : rbend,l_arc = larc, angle=ang
bb: b, g_err = -2*ang/larc
""",
            [
                Element(
                    name=T("b"),
                    keyword=T("rbend"),
                    ele_list=None,
                    attributes=[
                        Attribute(name=T("l_arc"), value=T("larc")),
                        Attribute(name=T("angle"), value=T("ang")),
                    ],
                ),
                Element(
                    name=T("bb"),
                    keyword=T("b"),
                    ele_list=None,
                    attributes=[
                        Attribute(
                            name=T("g_err"),
                            value=Seq(
                                items=[
                                    T("-2"),
                                    STAR,
                                    T("ang"),
                                    SLASH,
                                    T("larc"),
                                ],
                                delimiter=SPACE,
                            ),
                        )
                    ],
                ),
            ],
            id="element_with_error",
        ),
        pytest.param(
            """
D_PATCH: patch, x_offset = 0.034, x_pitch = asin(0.32)
""",
            [
                Element(
                    name=T("D_PATCH"),
                    keyword=T("patch"),
                    ele_list=None,
                    attributes=[
                        Attribute(name=T("x_offset"), value=T("0.034")),
                        Attribute(
                            name=T("x_pitch"),
                            value=Seq(
                                items=[
                                    T("asin"),
                                    Seq(
                                        opener=LPAREN,
                                        closer=RPAREN,
                                        items=[T("0.32")],
                                        delimiter=COMMA,
                                    ),
                                ],
                                delimiter=SPACE,
                            ),
                        ),
                    ],
                )
            ],
            id="patch_with_expression",
        ),
        pytest.param(
            """
RETURN1.TIME_PATCH[T_OFFSET] = +1.22459198215588780E-10
RETURN7.TIME_PATCH[T_OFFSET] = RETURN1.TIME_PATCH[T_OFFSET]
""",
            [
                Parameter(
                    target=T("RETURN1.TIME_PATCH"),
                    name=T("T_OFFSET"),
                    value=T("+1.22459198215588780E-10"),
                ),
                Parameter(
                    target=T("RETURN7.TIME_PATCH"),
                    name=T("T_OFFSET"),
                    value=Seq(
                        items=[
                            T("RETURN1.TIME_PATCH"),
                            Seq(
                                opener=LBRACK,
                                closer=RBRACK,
                                items=[T("T_OFFSET")],
                            ),
                        ],
                    ),
                ),
            ],
            id="element_attribute_reference",
        ),
        pytest.param(
            """
RETURN1.TIME_MATCH[DELTA_TIME] =  -8.273e-12
RETURN5.TIME_MATCH[DELTA_TIME] =  return3.time_match[delta_time]
""",
            [
                Parameter(
                    target=T("RETURN1.TIME_MATCH"),
                    name=T("DELTA_TIME"),
                    value=T("-8.273e-12"),
                ),
                Parameter(
                    target=T("RETURN5.TIME_MATCH"),
                    name=T("DELTA_TIME"),
                    value=Seq(
                        items=[
                            T("return3.time_match"),
                            Seq(
                                opener=LBRACK,
                                closer=RBRACK,
                                items=[T("delta_time")],
                            ),
                        ],
                    ),
                ),
            ],
            id="lowercase_attribute_reference",
        ),
        pytest.param(
            """
cavity7[wall] = {
section = { s = 0., v(1) = {0.0, 0.0, 0.05519118826235362}},
}
""",
            [
                Parameter(
                    target=T("cavity7"),
                    name=T("wall"),
                    value=Seq(
                        opener=LBRACE,
                        closer=RBRACE,
                        delimiter=COMMA,
                        items=[
                            Attribute(
                                name=T("section"),
                                value=Seq(
                                    opener=LBRACE,
                                    closer=RBRACE,
                                    delimiter=COMMA,
                                    items=[
                                        Attribute(name=T("s"), value=T("0.")),
                                        Attribute(
                                            name=CallName(
                                                name=T("v"),
                                                args=make_seq(
                                                    opener="(", items=[T("1")], delimiter=COMMA
                                                ),
                                            ),
                                            value=Seq(
                                                opener=LBRACE,
                                                closer=RBRACE,
                                                items=[
                                                    T("0.0"),
                                                    T("0.0"),
                                                    T("0.05519118826235362"),
                                                ],
                                                delimiter=COMMA,
                                            ),
                                        ),
                                    ],
                                ),
                            )
                        ],
                    ),
                )
            ],
            id="wall_multiline_definition",
        ),
        pytest.param(
            """
IN.CRMOD.Pip01: PIPE, L = 5.318*0.0254, aperture = 0.060198/2,
    wall = {
	section = { s = 0,
	  v(1) = {0, 0, IN.CRMOD.Pip01[aperture]}},
	section = { s = IN.CRMOD.Pip01[L],
	  v(1) = {0, 0, IN.CRMOD.Pip01[aperture]}}}
""",
            [
                Element(
                    name=T("IN.CRMOD.Pip01"),
                    keyword=T("PIPE"),
                    ele_list=None,
                    attributes=[
                        Attribute(
                            name=T("L"),
                            value=Seq(
                                items=[
                                    T("5.318"),
                                    STAR,
                                    T("0.0254"),
                                ]
                            ),
                        ),
                        Attribute(
                            name=T("aperture"),
                            value=Seq(
                                items=[
                                    T("0.060198"),
                                    SLASH,
                                    T("2"),
                                ]
                            ),
                        ),
                        Attribute(
                            name=T("wall"),
                            value=Seq(
                                opener=LBRACE,
                                closer=RBRACE,
                                delimiter=COMMA,
                                items=[
                                    Attribute(
                                        name=T("section"),
                                        value=Seq(
                                            opener=LBRACE,
                                            closer=RBRACE,
                                            delimiter=COMMA,
                                            items=[
                                                Attribute(name=T("s"), value=T("0")),
                                                Attribute(
                                                    name=CallName(
                                                        name=T("v"),
                                                        args=make_seq(
                                                            opener="(",
                                                            items=[T("1")],
                                                            delimiter=COMMA,
                                                        ),
                                                    ),
                                                    value=Seq(
                                                        opener=LBRACE,
                                                        closer=RBRACE,
                                                        delimiter=COMMA,
                                                        items=[
                                                            T("0"),
                                                            T("0"),
                                                            Seq(
                                                                items=[
                                                                    T("IN.CRMOD.Pip01"),
                                                                    Seq(
                                                                        opener=LBRACK,
                                                                        closer=RBRACK,
                                                                        items=[T("aperture")],
                                                                    ),
                                                                ],
                                                            ),
                                                        ],
                                                    ),
                                                ),
                                            ],
                                        ),
                                    ),
                                    Attribute(
                                        name=T("section"),
                                        value=Seq(
                                            opener=LBRACE,
                                            closer=RBRACE,
                                            delimiter=COMMA,
                                            items=[
                                                Attribute(
                                                    name=T("s"),
                                                    value=Seq(
                                                        items=[
                                                            T("IN.CRMOD.Pip01"),
                                                            Seq(
                                                                opener=LBRACK,
                                                                closer=RBRACK,
                                                                items=[T("L")],
                                                            ),
                                                        ],
                                                    ),
                                                ),
                                                Attribute(
                                                    name=CallName(
                                                        name=T("v"),
                                                        args=make_seq(
                                                            opener="(",
                                                            items=[T("1")],
                                                            delimiter=COMMA,
                                                        ),
                                                    ),
                                                    value=Seq(
                                                        opener=LBRACE,
                                                        closer=RBRACE,
                                                        delimiter=COMMA,
                                                        items=[
                                                            T("0"),
                                                            T("0"),
                                                            Seq(
                                                                items=[
                                                                    T("IN.CRMOD.Pip01"),
                                                                    Seq(
                                                                        opener=LBRACK,
                                                                        closer=RBRACK,
                                                                        items=[T("aperture")],
                                                                    ),
                                                                ],
                                                            ),
                                                        ],
                                                    ),
                                                ),
                                            ],
                                        ),
                                    ),
                                ],
                            ),
                        ),
                    ],
                )
            ],
            id="complex_wall_definition",
        ),
        pytest.param(
            """
facT : overlay = {QTov1, QTov2, QTov3, QTov4, QTov5, QTov6, QTov7, QTov8}, var={dk}, dk=1.0193
""",
            [
                Element(
                    name=T("facT"),
                    keyword=T("overlay"),
                    ele_list=Seq(
                        opener=LBRACE,
                        closer=RBRACE,
                        delimiter=COMMA,
                        items=[
                            T("QTov1"),
                            T("QTov2"),
                            T("QTov3"),
                            T("QTov4"),
                            T("QTov5"),
                            T("QTov6"),
                            T("QTov7"),
                            T("QTov8"),
                        ],
                    ),
                    attributes=[
                        Attribute(
                            name=T("var"),
                            value=Seq(
                                opener=LBRACE,
                                closer=RBRACE,
                                items=[T("dk")],
                                delimiter=COMMA,
                            ),
                        ),
                        Attribute(name=T("dk"), value=T("1.0193")),
                    ],
                )
            ],
            id="overlay_multiple_masters",
        ),
        pytest.param(
            """
tmatch : match, beta_a0=1, beta_a1=1, beta_b0=1, beta_b1=1
return1.time_match: tmatch
""",
            [
                Element(
                    name=T("tmatch"),
                    keyword=T("match"),
                    ele_list=None,
                    attributes=[
                        Attribute(name=T("beta_a0"), value=T("1")),
                        Attribute(name=T("beta_a1"), value=T("1")),
                        Attribute(name=T("beta_b0"), value=T("1")),
                        Attribute(name=T("beta_b1"), value=T("1")),
                    ],
                ),
                Element(name=T("return1.time_match"), keyword=T("tmatch")),
            ],
            id="match_element_parameters",
        ),
        pytest.param(
            """
A_PATCH: patch, flexible = T
""",
            Element(
                name=T("A_PATCH"),
                keyword=T("patch"),
                ele_list=None,
                attributes=[Attribute(name=T("flexible"), value=T("T"))],
            ),
            id="flexible_patch",
        ),
        pytest.param(
            """
*[tracking_method] = bmad_standard  ! Matches all elements.
""",
            [
                Parameter(
                    comments=Comments(inline=T(" Matches all elements.")),
                    target=STAR,
                    name=T("tracking_method"),
                    value=T("bmad_standard"),
                )
            ],
            id="wildcard_all",
        ),
        pytest.param(
            """
Q%1[k1] = 0.234                     ! Matches to "Q01" but not "Q001".
""",
            [
                Parameter(
                    comments=Comments(inline=T(' Matches to "Q01" but not "Q001".')),
                    target=T("Q%1"),
                    name=T("k1"),
                    value=T("0.234"),
                )
            ],
            id="single_character_match",
        ),
        pytest.param(
            """
quadrupole::Q*[k1] = 0.234    ! Matches all quadrupoles with names beginning with Q.
""",
            [
                Parameter(
                    comments=Comments(
                        inline=T(" Matches all quadrupoles with names beginning with Q.")
                    ),
                    target=Seq(
                        delimiter=SPACE,
                        items=[T("quadrupole"), Delimiter("::"), T("Q"), STAR],
                    ),
                    name=T("k1"),
                    value=T("0.234"),
                )
            ],
            id="class_qualified_wildcard",
        ),
        pytest.param(
            """
expand_lattice              ! Expand the lattice.
97[x_offset] = 0.0023       ! Set x_offset attribute of 97th element
b2>>si_cryst##2[tilt] = 0.1 ! Tilt the 2nd instance of "si_cryst" in branch "b2"
5:32[x_limit] = 0.3         ! Sets elements with indexes 5 through 32 in branch 0.
""",
            [
                Simple(
                    comments=Comments(inline=T(" Expand the lattice.")),
                    statement=T("expand_lattice"),
                    arguments=[],
                ),
                Parameter(
                    comments=Comments(inline=T(" Set x_offset attribute of 97th element")),
                    target=T("97"),
                    name=T("x_offset"),
                    value=T("0.0023"),
                ),
                Parameter(
                    comments=Comments(
                        inline=T(' Tilt the 2nd instance of "si_cryst" in branch "b2"')
                    ),
                    target=T("b2>>si_cryst##2"),
                    name=T("tilt"),
                    value=T("0.1"),
                ),
                Parameter(
                    comments=Comments(
                        inline=T(" Sets elements with indexes 5 through 32 in branch 0.")
                    ),
                    target=Seq(items=[T("5"), COLON, T("32")]),
                    name=T("x_limit"),
                    value=T("0.3"),
                ),
            ],
            id="element_index_syntax",
        ),
        pytest.param(
            """
2>>45[x_pitch] = 0.1
x_br>>quad::q*[k1] = 0.5
""",
            [
                Parameter(target=T("2>>45"), name=T("x_pitch"), value=T("0.1")),
                Parameter(
                    target=Seq(
                        delimiter=SPACE,
                        items=[T("x_br>>quad"), Delimiter("::"), T("q"), STAR],
                    ),
                    name=T("k1"),
                    value=T("0.5"),
                ),
            ],
            id="branch_qualified_name",
        ),
        pytest.param(
            """
sbend::q1:q5[field_master] = T
""",
            [
                Parameter(
                    target=Seq(
                        delimiter=SPACE,
                        items=[
                            T("sbend"),
                            Delimiter("::"),
                            T("q1"),
                            COLON,
                            T("q5"),
                        ],
                    ),
                    name=T("field_master"),
                    value=T("T"),
                )
            ],
            id="element_range_syntax_between_named_eles",
        ),
        pytest.param(
            """
3,15:17[x_offset] = 0.001
""",
            [
                Parameter(
                    target=Seq(
                        delimiter=COMMA,
                        items=[T("3"), Seq(items=[T("15"), COLON, T("17")])],
                    ),
                    name=T("x_offset"),
                    value=T("0.001"),
                )
            ],
            id="element_range_syntax_with_commas",
        ),
        pytest.param(
            """
3 15:17[x_offset] = 0.001
""",
            [
                Parameter(
                    target=Seq(
                        items=[T("3"), T("15"), COLON, T("17")],
                    ),
                    name=T("x_offset"),
                    value=T("0.001"),
                )
            ],
            id="element_range_syntax_without_commas",
        ),
        pytest.param(
            """
100:200 & sbend::*[tracking_method] = runge_kutta
""",
            [
                Parameter(
                    target=Seq(
                        items=[
                            T("100"),
                            COLON,
                            T("200"),
                            AMPERSAND,
                            T("sbend"),
                            Delimiter("::"),
                            T("*"),
                        ],
                    ),
                    name=T("tracking_method"),
                    value=T("runge_kutta"),
                )
            ],
            id="list_intersection",
        ),
        pytest.param(
            """
q* & quad::* b1[k1] = 0.5
""",
            [
                Parameter(
                    target=Seq(
                        items=[
                            T("q"),
                            STAR,
                            AMPERSAND,
                            T("quad"),
                            Delimiter("::"),
                            T("*"),
                            T("b1"),
                        ],
                    ),
                    name=T("k1"),
                    value=T("0.5"),
                )
            ],
            id="list_intersection_three",
        ),
        pytest.param(
            """
*::*, ~quadrupole::*[x_offset] = 0.001  ! All elements except quadrupoles
""",
            [
                Parameter(
                    comments=Comments(inline=T(" All elements except quadrupoles")),
                    target=Seq(
                        delimiter=COMMA,
                        items=[
                            Seq(delimiter=SPACE, items=[T("*"), Delimiter("::"), T("*")]),
                            Seq(delimiter=SPACE, items=[T("~quadrupole"), Delimiter("::"), T("*")]),
                        ],
                    ),
                    name=T("x_offset"),
                    value=T("0.001"),
                )
            ],
            id="list_exclusion",
        ),
        pytest.param(
            """
! All elements with names beginning with B except elements with names beginning
! with B1. However, elements named B13 are retained.
b*, ~b1*, b13[hkick] = 0.01
""",
            [
                Parameter(
                    comments=Comments(
                        pre=[
                            T(
                                " All elements with names beginning with B except elements with names beginning"
                            ),
                            T(" with B1. However, elements named B13 are retained."),
                        ]
                    ),
                    target=Seq(
                        delimiter=COMMA,
                        items=[
                            Seq(delimiter=SPACE, items=[T("b"), STAR]),
                            Seq(delimiter=SPACE, items=[T("~b1"), STAR]),
                            T("b13"),
                        ],
                    ),
                    name=T("hkick"),
                    value=T("0.01"),
                )
            ],
            id="list_exclusion1",
        ),
        pytest.param(
            """
type::"det bpm*"[x_pitch] = 0.1  ! All elements with a type starting with "det bpm"
alias::*[is_on] = F  ! all elements with a non-blank alias
""",
            [
                Parameter(
                    comments=Comments(
                        inline=T(' All elements with a type starting with "det bpm"')
                    ),
                    target=Seq(
                        delimiter=SPACE, items=[T("type"), Delimiter("::"), T('"det bpm*"')]
                    ),
                    name=T("x_pitch"),
                    value=T("0.1"),
                ),
                Parameter(
                    comments=Comments(inline=T(" all elements with a non-blank alias")),
                    target=Seq(delimiter=SPACE, items=[T("alias"), Delimiter("::"), T("*")]),
                    name=T("is_on"),
                    value=T("F"),
                ),
            ],
            id="attribute_matching",
        ),
        pytest.param(
            """
s%z[k2] = %[k2] + 0.03 * ran_gauss()  ! %[k2] refers to the element being set (like s0z)
""",
            [
                Parameter(
                    comments=Comments(
                        inline=T(" %[k2] refers to the element being set (like s0z)")
                    ),
                    target=T("s%z"),
                    name=T("k2"),
                    value=Seq(
                        items=[
                            T("%"),
                            Seq(
                                opener=LBRACK,
                                closer=RBRACK,
                                items=[T("k2")],
                                delimiter=SPACE,
                            ),
                            T("+"),
                            T("0.03"),
                            T("*"),
                            T("ran_gauss"),
                            Seq(opener=LPAREN, closer=RPAREN, items=[], delimiter=COMMA),
                        ],
                    ),
                )
            ],
            id="percent_in_expression",
        ),
        pytest.param(
            """
c: crystal, call::$AB/my_curvature.bmad, h_misalign = call::my_surface.bmad
""",
            [
                Element(
                    name=T("c"),
                    keyword=T("crystal"),
                    ele_list=None,
                    attributes=[
                        Attribute(
                            name=Seq(
                                delimiter=SPACE,
                                items=[
                                    T("call"),
                                    Delimiter("::"),
                                    T("$AB"),
                                    SLASH,
                                    T("my_curvature.bmad"),
                                ],
                            ),
                            value=None,
                        ),
                        Attribute(
                            name=T("h_misalign"),
                            value=Seq(
                                delimiter=SPACE,
                                items=[T("call"), Delimiter("::"), T("my_surface.bmad")],
                            ),
                        ),
                    ],
                )
            ],
            id="inline_call",
        ),
        pytest.param(
            "parameter[particle] = deuteron",
            [Parameter(target=T("parameter"), name=T("particle"), value=T("deuteron"))],
            id="species_function_0",
        ),
        pytest.param(
            "am = anomalous_moment_of(parameter[particle])^2",
            [
                Constant(
                    name=T("am"),
                    value=Seq(
                        delimiter=SPACE,
                        items=[
                            T("anomalous_moment_of"),
                            Seq(
                                opener=LPAREN,
                                closer=RPAREN,
                                delimiter=COMMA,
                                items=[
                                    Seq(
                                        delimiter=SPACE,
                                        items=[
                                            T("parameter"),
                                            Seq(
                                                opener=LBRACK,
                                                closer=RBRACK,
                                                items=[T("particle")],
                                                delimiter=SPACE,
                                            ),
                                        ],
                                    )
                                ],
                                # items=[T("parameter[particle]")],
                            ),
                            T("^"),
                            T("2"),
                        ],
                    ),
                    redef=False,
                )
            ],
            id="species_function_1",
        ),
        pytest.param(
            "my_particle = species(He++)      ! my_particle now represents He++",
            [
                Constant(
                    comments=Comments(inline=T(" my_particle now represents He++")),
                    name=T("my_particle"),
                    value=Seq(
                        items=[
                            T("species"),
                            Seq(
                                opener=LPAREN,
                                closer=RPAREN,
                                items=[T("He++")],
                                delimiter=COMMA,
                            ),
                        ],
                    ),
                    redef=False,
                )
            ],
            id="species_function_2",
        ),
        pytest.param(
            "chg1 = charge_of(my_particle)    ! chg = charge of He++",
            [
                Constant(
                    comments=Comments(inline=T(" chg = charge of He++")),
                    name=T("chg1"),
                    value=Seq(
                        items=[
                            T("charge_of"),
                            Seq(
                                opener=LPAREN,
                                closer=RPAREN,
                                items=[T("my_particle")],
                                delimiter=COMMA,
                            ),
                        ],
                    ),
                    redef=False,
                )
            ],
            id="species_function_3",
        ),
        pytest.param(
            "chg2 = charge_of(He++)           ! Same as previous line",
            [
                Constant(
                    comments=Comments(inline=T(" Same as previous line")),
                    name=T("chg2"),
                    value=Seq(
                        items=[
                            T("charge_of"),
                            Seq(
                                opener=LPAREN,
                                closer=RPAREN,
                                items=[T("He++")],
                                delimiter=COMMA,
                            ),
                        ],
                    ),
                    redef=False,
                )
            ],
            id="species_function_4",
        ),
        pytest.param(
            "chg3 = charge_of(species(He++))  ! Same as previous line",
            [
                Constant(
                    comments=Comments(inline=T(" Same as previous line")),
                    name=T("chg3"),
                    value=Seq(
                        items=[
                            T("charge_of"),
                            Seq(
                                opener=LPAREN,
                                closer=RPAREN,
                                delimiter=COMMA,
                                items=[
                                    Seq(
                                        delimiter=SPACE,
                                        items=[
                                            T("species"),
                                            Seq(
                                                opener=LPAREN,
                                                closer=RPAREN,
                                                items=[T("He++")],
                                                delimiter=COMMA,
                                            ),
                                        ],
                                    )
                                ],
                            ),
                        ],
                    ),
                    redef=False,
                )
            ],
            id="species_function_5",
        ),
        pytest.param(
            """
a[x_offset] = 0.001*ran_gauss(3.0)
b[y_offset] = 0.002*ran_gauss()
""",
            [
                Parameter(
                    target=T("a"),
                    name=T("x_offset"),
                    value=Seq(
                        items=[
                            T("0.001"),
                            STAR,
                            T("ran_gauss"),
                            Seq(
                                opener=LPAREN,
                                closer=RPAREN,
                                items=[T("3.0")],
                                delimiter=COMMA,
                            ),
                        ],
                    ),
                ),
                Parameter(
                    target=T("b"),
                    name=T("y_offset"),
                    value=Seq(
                        items=[
                            T("0.002"),
                            STAR,
                            T("ran_gauss"),
                            Seq(opener=LPAREN, closer=RPAREN, items=[], delimiter=COMMA),
                        ],
                    ),
                ),
            ],
            id="ran_gauss_with_cut",
        ),
        pytest.param(
            """
parameter[E_TOT] = 5e9
parameter[geometry] = open
parameter[particle] = POSITRON
parameter[ran_seed] = 12345
""",
            [
                Parameter(target=T("parameter"), name=T("E_TOT"), value=T("5e9")),
                Parameter(target=T("parameter"), name=T("geometry"), value=T("open")),
                Parameter(target=T("parameter"), name=T("particle"), value=T("POSITRON")),
                Parameter(target=T("parameter"), name=T("ran_seed"), value=T("12345")),
            ],
            id="docs_parameter_statement",
        ),
        pytest.param(
            """
beginning[beta_a] = 14.5011548
beginning[alpha_a] = -0.53828197
beginning[theta_position] = theta_in + phi/2
""",
            [
                Parameter(target=T("beginning"), name=T("beta_a"), value=T("14.5011548")),
                Parameter(target=T("beginning"), name=T("alpha_a"), value=T("-0.53828197")),
                Parameter(
                    target=T("beginning"),
                    name=T("theta_position"),
                    value=Seq(
                        items=[
                            T("theta_in"),
                            PLUS,
                            T("phi"),
                            SLASH,
                            T("2"),
                        ]
                    ),
                ),
            ],
            id="docs_beginning_statement",
        ),
        pytest.param(
            """
qq: quadrupole, r_custom(-2,1,5) = 34.5, r_custom(-3) = 77.9
""",
            [
                Element(
                    name=T("qq"),
                    keyword=T("quadrupole"),
                    ele_list=None,
                    attributes=[
                        Attribute(
                            name=CallName(
                                name=T("r_custom"),
                                args=make_seq(
                                    opener="(", items=[T("-2"), T("1"), T("5")], delimiter=COMMA
                                ),
                            ),
                            value=T("34.5"),
                        ),
                        Attribute(
                            name=CallName(
                                name=T("r_custom"),
                                args=make_seq(opener="(", items=[T("-3")], delimiter=COMMA),
                            ),
                            value=T("77.9"),
                        ),
                    ],
                )
            ],
            id="r_custom_array",
        ),
        pytest.param(
            """
qa, quad, l = 0.6, tilt = pi/4  ! Define QA.
qb: qa                          ! QB Inherits from QA.
qa[k1] = 0.12                   ! QB unaffected by modifications of QA.
""",
            [
                Assignment(
                    comments=Comments(inline=T(" Define QA.")),
                    name=Seq(
                        delimiter=COMMA,
                        items=[T("qa"), T("quad"), T("l")],
                    ),
                    value=Seq(
                        delimiter=COMMA,
                        items=[
                            T("0.6"),
                            Attribute(
                                name=T("tilt"),
                                value=Seq(delimiter=SPACE, items=[T("pi"), T("/"), T("4")]),
                            ),
                        ],
                    ),
                ),
                Element(
                    comments=Comments(inline=T(" QB Inherits from QA.")),
                    name=T("qb"),
                    keyword=T("qa"),
                ),
                Parameter(
                    comments=Comments(inline=T(" QB unaffected by modifications of QA.")),
                    target=T("qa"),
                    name=T("k1"),
                    value=T("0.12"),
                ),
            ],
            id="element_inheritance",
        ),
        pytest.param(
            """
BR: fork, to_line = RING_L, to_element = M
""",
            [
                Element(
                    name=T("BR"),
                    keyword=T("fork"),
                    ele_list=None,
                    attributes=[
                        Attribute(name=T("to_line"), value=T("RING_L")),
                        Attribute(name=T("to_element"), value=T("M")),
                    ],
                )
            ],
            id="fork_element",
        ),
        pytest.param(
            """
parser_debug var lat ele 34 78
""",
            [
                Simple(
                    statement=T("parser_debug"),
                    arguments=[
                        T("var"),
                        T("lat"),
                        T("ele"),
                        T("34"),
                        T("78"),
                    ],
                )
            ],
            id="parser_debug",
        ),
        pytest.param(
            """
print "Remember Q01 quad strength of `q01[k1]` not yet optimized"
""",
            [
                Simple(
                    statement=T("print"),
                    arguments=[T('"Remember Q01 quad strength of `q01[k1]` not yet optimized"')],
                )
            ],
            id="print_with_quotes",
        ),
        pytest.param(
            """
call, no_abort_on_open_error, file = "file a b c"
""",
            [
                Simple(
                    statement=T("call"),
                    arguments=[
                        Attribute(name=T("no_abort_on_open_error"), value=None),
                        Attribute(name=T("file"), value=T('"file a b c"')),
                    ],
                )
            ],
            id="call_no_abort_on_error",
        ),
        pytest.param(
            """
call , f = "file a b c"
""",
            [
                Simple(
                    statement=T("call"),
                    arguments=[
                        Attribute(name=T("f"), value=T('"file a b c"')),
                    ],
                )
            ],
            id="call_regular",
        ),
        pytest.param(
            """
beam, particle = electron, n_part = 1.6e10
""",
            [
                Simple(
                    statement=T("beam"),
                    arguments=[
                        Attribute(name=T("particle"), value=T("electron")),
                        Attribute(name=T("n_part"), value=T("1.6e10")),
                    ],
                )
            ],
            id="beam_statement",
        ),
        pytest.param(
            """
particle_start[x]    = 0.2
""",
            [Parameter(target=T("particle_start"), name=T("x"), value=T("0.2"))],
            id="particle_start",
        ),
        pytest.param(
            """
title, "foo"
""",
            [Simple(statement=T("title"), arguments=[Attribute(name=T('"foo"'))])],
            id="title_oneline",
        ),
        pytest.param(
            """
title
"foo"
""",
            [Simple(statement=T("title"), arguments=[Attribute(name=T('"foo"'))])],
            id="title_twoline",
        ),
    ],
)


@code_tests
def test_parsing(code: str, expected) -> None:
    res = parse_verbose(code)

    if not isinstance(expected, list):
        (res,) = res

    assert res == expected


@code_tests
def test_roundtrip(code: str, expected) -> None:
    roundtrip_code(code)


@pytest.mark.parametrize(
    ("code",),
    [
        pytest.param("ename[amp_vs_time(1)%time]"),
        pytest.param("ename[amp_vs_time(N)%amp]"),  #
        pytest.param("ename[frequencies(N)%freq]"),
        pytest.param("ename[frequencies(N)%amp]"),  #
        pytest.param("ename[frequencies(N)%phi]"),  #
        pytest.param("ename[cartesian_map(N)%field_scale]"),  #
        pytest.param("ename[cartesian_map(N)%r0(1)]"),
        pytest.param("ename[cartesian_map(N)%r0(2)]"),
        pytest.param("ename[cartesian_map(N)%r0(3)]"),
        pytest.param("ename[cartesian_map(N)%master_parameter]"),
        pytest.param("ename[cartesian_map(N)%t(M)%A]"),  # -- M^th term in N^th map.
        pytest.param("ename[cartesian_map(N)%t(M)%kx]"),
        pytest.param("ename[cartesian_map(N)%t(M)%ky]"),
        pytest.param("ename[cartesian_map(N)%t(M)%kz]"),
        pytest.param("ename[cartesian_map(N)%t(M)%x0]"),
        pytest.param("ename[cartesian_map(N)%t(M)%y0]"),
        pytest.param("ename[cartesian_map(N)%t(M)%phi_z]"),
        pytest.param("ename[x_knot(N)]"),  # -- N^th x_knot point.
        pytest.param("ename[slave(M)%y_knot(N)]"),  # -- N^th y_knot point for M^th slave.
        pytest.param("ename[r_custom(N1, N2, N3)]"),
        pytest.param(
            "ename[r_custom(N1,N2)]"
        ),  # N2)]                 -- Equivalent to ename[r_custom(N1, N2,  0)]
        pytest.param("ename[r_custom(N1)]"),  # -- Equivalent to ename[r_custom(N1,  0,  0)]
        pytest.param("ename[cylindrical_map(N)%phi0_fieldmap]"),  #
        pytest.param("ename[cylindrical_map(N)%theta0_azimuth]"),
        pytest.param("ename[cylindrical_map(N)%field_scale]"),
        pytest.param("ename[cylindrical_map(N)%dz]"),
        pytest.param("ename[cylindrical_map(N)%r0(1)]"),
        pytest.param("ename[cylindrical_map(N)%r0(2)]"),
        pytest.param("ename[cylindrical_map(N)%r0(3)]"),
        pytest.param("ename[cylindrical_map(N)%master_parameter]"),
        pytest.param("ename[gen_grad_map(N)%field_scale]"),  #
        pytest.param("ename[gen_grad_map(N)%r0(1)]"),
        pytest.param("ename[gen_grad_map(N)%r0(2)]"),
        pytest.param("ename[gen_grad_map(N)%r0(3)]"),
        pytest.param("ename[gen_grad_map(N)%master_parameter]"),
        pytest.param("ename[grid_field(N)%phi0_fieldmap]"),  #
        pytest.param("ename[grid_field(N)%interpolation_order]"),
        pytest.param("ename[grid_field(N)%harmonic]"),
        pytest.param("ename[grid_field(N)%geometry]"),
        pytest.param("ename[grid_field(N)%ename_anchor_pt]"),
        pytest.param("ename[grid_field(N)%phi0_fieldmap]"),
        pytest.param("ename[grid_field(N)%field_scale]"),
        pytest.param("ename[grid_field(N)%dr(1)]"),
        pytest.param("ename[grid_field(N)%dr(2)]"),
        pytest.param("ename[grid_field(N)%dr(3)]"),
        pytest.param("ename[grid_field(N)%r0(1)]"),
        pytest.param("ename[grid_field(N)%r0(2)]"),
        pytest.param("ename[grid_field(N)%r0(3)]"),
        pytest.param("ename[grid_field(N)%master_parameter]"),
        pytest.param("ename[lr_wake%amp_scale]"),
        pytest.param("ename[lr_wake%time_scale]"),
        pytest.param("ename[lr_wake%freq_spread]"),
        pytest.param("ename[lr_wake%mode(N)%freq_in]"),  #
        pytest.param("ename[lr_wake%mode(N)%freq]"),  #
        pytest.param("ename[lr_wake%mode(N)%r_over_q]"),  #
        pytest.param("ename[lr_wake%mode(N)%damp]"),  #
        pytest.param("ename[lr_wake%mode(N)%phi]"),  #
        pytest.param("ename[lr_wake%mode(N)%polar_angle]"),
        pytest.param("ename[lr_wake%mode(N)%polarized]"),  #
        pytest.param("ename[curvature%spherical]"),
        pytest.param("ename[curvature%elliptical_x]"),
        pytest.param("ename[curvature%elliptical_y]"),
        pytest.param("ename[curvature%elliptical_z]"),
        pytest.param("ename[curvature%x(N1)y(N2)]"),
        pytest.param("ename[tt012]"),  # ! Orbital terms. EG: rot[tt13] -> M13 matrix term
        pytest.param("ename[ttS0123]"),  # ! S0 spin quaternion terms.
        pytest.param("ename[ttSx123]"),  # ! Sx spin quaternion terms.
        pytest.param("ename[ttSy123]"),  # ! Sy spin quaternion terms.
        pytest.param("ename[ttSz123]"),  # ! Sz spin quaternion terms.
        pytest.param("ename[wall%section(N)%s]"),
        pytest.param("ename[wall%section(N)%wall%dr_ds]"),
        pytest.param("ename[wall%section(N)%v(M)%x]"),
        pytest.param("ename[wall%section(N)%v(M)%y]"),
        pytest.param("ename[wall%section(N)%v(M)%radius_x]"),
        pytest.param("ename[wall%section(N)%v(M)%radius_y]"),
        pytest.param("ename[wall%section(N)%v(M)%tilt]"),
    ],
)
def test_nonstandard_parameter_syntax(code: str) -> None:
    target, rest = code.split("[", 1)
    name = rest.rstrip("]")

    expected = NonstandardParameter(
        target=T(target),
        name=T(name),
        value=T("0"),
    )

    code = f"{code} = 0"
    (res,) = parse_verbose(code)

    if not isinstance(res, NonstandardParameter):
        expected = Parameter(
            target=T(target),
            name=T(name),
            value=T("0"),
        )
        # TODO this is aliased to regular Parameter; is it an issue?

    if expected.name == "r_custom(N1, N2, N3)":
        # TODO note spacing
        expected.name = T("r_custom(N1,N2,N3)")

    assert res == expected

    # These aren't in the parametrized list, don't overlook them
    roundtrip_code(code)


def test_attribute_in_parameter_assignment() -> None:
    code = """
foo[bar] = bar = baz
"""
    with pytest.raises(UnexpectedAssignment):
        parse_verbose(code)


def test_attribute_in_constant_target() -> None:
    code = """
foo = bar = baz
"""
    with pytest.raises(UnexpectedAssignment):
        parse_verbose(code)


@pytest.mark.parametrize(
    "code",
    [
        pytest.param(
            """
            foo = '
            '
            """,
            id="line-break-squote",
        ),
        pytest.param(
            """
            foo = "
            "
            """,
            id="line-break-dquote",
        ),
        pytest.param(
            "'",
            id="single-line-squote",
        ),
        pytest.param(
            '"',
            id="single-line-dquote",
        ),
        pytest.param(
            "'foo",
            id="single-line-squote",
        ),
        pytest.param(
            '"foo',
            id="single-line-dquote",
        ),
        pytest.param(
            "foo'",
            id="single-line-squote",
        ),
        pytest.param(
            'foo"',
            id="single-line-dquote",
        ),
    ],
)
def test_unterminated_string(code: str) -> None:
    with pytest.raises(UnterminatedString):
        parse_verbose(code)


@pytest.mark.parametrize(
    "code",
    [
        pytest.param(
            """
            "'''"
            """,
            id="squote-in-dquote",
        ),
        pytest.param(
            """
            '""'
            """,
            id="dquote-in-squote",
        ),
    ],
)
def test_string_quote_nesting(code: str) -> None:
    parse_verbose(code)


@pytest.mark.parametrize(
    "string_",
    [
        pytest.param(
            "Remember Q01 quad strength of `q01[k1]` not yet optimized",
            id="remember",
        ),
        pytest.param(
            "Single quote 'string'",
            id="squote",
        ),
        pytest.param(
            'Double quote "string"',
            id="squote",
        ),
    ],
)
def test_print_unquoted(string_: str) -> None:
    code = f"""
print {string_}
"""
    expected = Simple(
        statement=T("print"),
        arguments=[T(string_)],
    )
    (res,) = parse_verbose(code)
    assert res == expected

    options = FormatOptions()
    roundtrip_code = format_statements(res, options)

    quoted = Token(string_).quoted()
    assert roundtrip_code == f"print {quoted}"



from __future__ import annotations

from ._attrs import by_element as element_key_to_attrs


def get_attributes_for_ele(ele_keyword: str):
    try:
        return element_key_to_attrs[ele_keyword.upper()]
    except KeyError:
        pass

    for key in element_key_to_attrs:
        if key.startswith(ele_keyword.upper()):
            return element_key_to_attrs[key]
    raise KeyError(f"Element keyword not found: {ele_keyword.upper()}")


def get_attribute(ele_keyword: str, name: str):
    attrs = get_attributes_for_ele(ele_keyword)
    return attrs[name.upper()]



from __future__ import annotations

import typing
from dataclasses import dataclass, field

if typing.TYPE_CHECKING:
    from .token import Token


@dataclass
class Comments:
    pre: list[Token] = field(default_factory=list)
    inline: Token | None = None

    def __bool__(self):
        return any((self.pre, self.inline))

    def clear(self) -> None:
        self.pre = []
        self.inline = None

    def clone(self) -> Comments:
        return Comments(pre=list(self.pre), inline=self.inline)

    def __repr__(self):
        parts = []
        if self.pre:
            parts.append(f"pre={self.pre}")
        if self.inline:
            parts.append(f"inline={self.inline!r}")

        return f"{type(self).__name__}({', '.join(parts)})"

    # def wrap_code(self, code: str, indent: str = "") -> str:
    #     out = ""
    #     for line in self.pre:
    #         out += f"{indent}! {line}\n"
    #     out += indent + code
    #     if self.inline:
    #         out += f"  ! {self.inline}\n"
    #     return out



from __future__ import annotations

import pathlib
from dataclasses import dataclass, field

from .const import DELIMITERS, LBRACE, LBRACK, LPAREN
from .exceptions import EndOfLine, MissingCloseDelimiter, UnterminatedString
from .funcs import ALL_BUILTIN
from .token import Comments, Delimiter, Location, Role, Token
from .types import Block
from .util import DelimiterState


@dataclass
class _StatementTokenBlock:
    """
    A raw block from the file that, when parsed, should correspond to a single statement.
    """

    lineno: int
    items: list[Token | Delimiter] = field(default_factory=list)

    def stack(self) -> Block:
        root = Block()
        stack = [root]
        delim_state = DelimiterState()
        for tok in self.items:
            if isinstance(tok, Delimiter):
                _, level_change = delim_state.update(tok)
                if level_change > 0:
                    new_block = Block(opener=tok)
                    stack[-1].items.append(new_block)
                    stack.append(new_block)
                elif level_change < 0:
                    last_block = stack.pop(-1)
                    last_block.closer = tok
                else:
                    stack[-1].items.append(tok)

            else:
                stack[-1].items.append(tok)

        if (
            not root.opener
            and not root.closer
            and len(root.items) == 1
            and isinstance(root.items[0], Block)
        ):
            return root.items[0]

        return root


def split_comment(
    line: str, filename: pathlib.Path | str, lineno: int = 0
) -> tuple[str, Token | None]:
    """
    Split code and comment at first '!' that is not inside quotes.
    """
    in_single = False
    in_double = False
    for idx, ch in enumerate(line):
        if ch == "'" and not in_double:
            in_single = not in_single
        elif ch == '"' and not in_single:
            in_double = not in_double
        elif ch == "!" and not in_single and not in_double:
            comment = line[idx + 1 :].rstrip()
            loc = Location(
                filename=pathlib.Path(filename),
                line=lineno,
                column=idx,
                end_line=lineno,
                end_column=idx + 1 + len(comment),
            )
            return line[:idx].rstrip(), Token(comment, loc=loc)
    return line.rstrip(), None


def _attach_inline_comment(
    first: Token | Block | None,
    last: Token | Block | None,
    inline: Token | None,
    pending: list[Token],
    is_multiline: bool,
) -> None:
    if first is None:
        if inline:
            pending.append(inline)
        return

    first.comments.pre.extend(pending)
    pending.clear()
    if inline is None:
        return
    if last is None:
        first.comments.inline = inline
        return

    if is_multiline or first.loc.line != last.loc.end_line:
        # On different lines; associate with the last token
        last.comments.inline = inline
    else:
        # On same line; associate with the first token
        first.comments.inline = inline


def _get_first_token_or_block(items) -> Token | Block | None:
    for item in items:
        if isinstance(item, (Token, Block)) and not isinstance(item, Delimiter):
            return item
    return None


_numeric_chars = frozenset({0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "."})


def _is_scientific_notation_start(tok: str | None) -> bool:
    if not tok:
        return False
    if not tok.lower().endswith("e"):
        return False
    val = tok[:-1]

    if _numeric_chars.issuperset(set(tok)):
        return False

    try:
        float(val)
    except ValueError:
        return False
    return True


class Tokenizer:
    filename: pathlib.Path
    line: str
    lineno: int
    pos: int

    def __init__(self, source: str, filename: pathlib.Path | str = "unknown") -> None:
        self.source = source
        self.filename = pathlib.Path(filename)

    def split_blocks(self):
        blocks: list[_StatementTokenBlock] = []
        last_delim = None
        delim_state = DelimiterState()
        pending_comments: list[Token] = []

        for self.lineno, line in enumerate(self.source.splitlines()):
            self.line, inline_comment = split_comment(
                line, lineno=self.lineno, filename=self.filename
            )
            self.pos = 0

            if not self.line.strip():
                if inline_comment is not None:
                    pending_comments.append(inline_comment)
                continue

            if last_delim == "&":
                block = blocks[-1]
                assert block.items[-1] == "&"
                block.items.pop(-1)
            elif last_delim in set("{[(,=") or delim_state.current_delimiters:
                block = blocks[-1]
            elif (
                # Dumb special case for multiline 'title' (see manual)
                blocks
                and blocks[-1].items
                and len(blocks[-1].items) == 1
                and isinstance(blocks[-1].items[0], Token)
                and blocks[-1].items[0].lower() == "title"
            ):
                block = blocks[-1]
            else:
                block = _StatementTokenBlock(lineno=self.lineno)
                blocks.append(block)

            line_start_idx = len(block.items)
            while True:
                try:
                    token, delim = self.get_next_word()
                except EndOfLine:
                    break

                if delim == ";":
                    # statement1; statement2
                    if token:
                        block.items.append(token)
                    last_delim = None
                    block = _StatementTokenBlock(lineno=self.lineno)
                    blocks.append(block)
                    continue

                elif last_delim in frozenset({"+", "-"}) and token and token[0].isnumeric():
                    # TODO : do these in a second pass
                    prev_item = block.items[-2] if len(block.items) > 1 else None
                    if (
                        prev_item in {"+", "-", "*", "/", "=", ",", LBRACE, LBRACK, LPAREN}
                        or prev_item is None
                    ):
                        token = Token.join([block.items.pop(), token])
                        last_delim = prev_item
                    block.items.append(token)
                elif token:
                    block.items.append(token)

                if delim:
                    delim_state.update(delim)
                    block.items.append(delim)

                last_delim = delim

            if not block.items:
                continue

            is_multiline = block.items[0].loc.line != block.items[-1].loc.line
            first_item = block.items[line_start_idx]
            last_item = (
                _get_first_token_or_block(
                    reversed(block.items[block.items.index(first_item) + 1 :])
                )
                if first_item
                else None
            )
            _attach_inline_comment(
                first_item, last_item, inline_comment, pending_comments, is_multiline
            )

            if block.items[0].lower() in ("return", "end_file"):
                # TODO: need to include whatever comes after these lines verbatim
                break

            if (
                block.items[0].lower() in ("print",)
                and len(block.items) > 1
                and not block.items[1].is_quoted_string
            ):
                block.items = [block.items[0], Token.join(block.items[1:])]

        if pending_comments:
            loc = sum((comment.loc for comment in pending_comments), pending_comments[0].loc)
            blocks.append(
                _StatementTokenBlock(
                    lineno=pending_comments[0].loc.line,
                    items=[
                        Token(
                            "",
                            loc=loc,
                            comments=Comments(pre=pending_comments),
                        ),
                    ],
                )
            )

        in_delims = delim_state.current_delimiters
        if in_delims:
            unmatched = ", ".join(f"{delim!r} at {delim.loc}" for delim in in_delims)
            raise MissingCloseDelimiter(f"Unmatched delimiters: {unmatched}", delims=in_delims)

        # Exclude any empty statements (semicolons at EOL, for example)
        return [block for block in blocks if block.items]

    def _scan_word(self, delimiters: frozenset[str]) -> tuple[str, int]:
        start = self.pos
        quote_pos = None
        quote_char = None
        while self.pos < len(self.line):
            ch = self.line[self.pos]
            if ch in ("'", '"'):
                if quote_pos is None:
                    quote_pos = self.pos
                    quote_char = ch
                elif ch == quote_char:
                    quote_pos = None
                    quote_char = None
                self.pos += 1
                continue
            if quote_pos is None and (ch in delimiters or ch.isspace()):
                break
            self.pos += 1

        if quote_pos is not None:
            loc = Location(self.filename, self.lineno, quote_pos, self.lineno, quote_pos)
            raise UnterminatedString(
                f"Unterminated quote character: {quote_char!r}",
                delim=Delimiter(quote_char, loc=loc),
            )

        word = self.line[start : self.pos]
        if (
            _is_scientific_notation_start(word)
            and self.pos < len(self.line)
            and self.line[self.pos] in {"-", "+"}
        ):
            plus_minus = self.line[self.pos]
            self.pos += 1
            end, end_pos = self._scan_word(delimiters)
            word = "".join((word, plus_minus, end))
            return word, end_pos

        return word, self.pos

    def _scan_delimiter(self, delimiters: frozenset[str]) -> Delimiter | None:
        while self.pos < len(self.line) and self.line[self.pos].isspace():
            self.pos += 1
        if self.pos >= len(self.line) or self.line[self.pos] not in delimiters:
            return None

        start = self.pos
        ch = self.line[self.pos]
        self.pos += 1

        # Handle compound delimiters and fake delimiters
        if self.pos < len(self.line):
            nxt = self.line[self.pos]

            match (ch, nxt):
                # Compound delimiters: :: and :=
                case (":", "=") | (":", ":"):
                    self.pos += 1
                    compound = ch + nxt
                    loc = Location(self.filename, self.lineno, start, self.lineno, self.pos)
                    if compound == ":=":
                        return Delimiter("=", loc)  # keep existing squashing
                    return Delimiter(compound, loc)

        loc = Location(self.filename, self.lineno, start, self.lineno, start + 1)
        return Delimiter(ch, loc)

    def get_next_word(
        self, delimiters: frozenset[str] = DELIMITERS, upper: bool = False
    ) -> tuple[Token | None, Delimiter | None]:
        """
        Get next word from the parse line.

        Parameters
        ----------
        delimiters : str, optional
            Valid delimiters for this context. If None, uses default.
        upper : bool, optional
            Convert word to uppercase.

        Returns
        -------
        word : str
            The extracted word
        delimiter : str
            The delimiter following the word
        """

        while self.pos < len(self.line) and self.line[self.pos].isspace():
            self.pos += 1

        if self.pos >= len(self.line):
            raise EndOfLine()
        # leading delimiter
        if self.line[self.pos] in delimiters:
            delim = self._scan_delimiter(delimiters)
            return None, delim

        word, end_pos = self._scan_word(delimiters)
        loc = Location(self.filename, self.lineno, end_pos - len(word), self.lineno, end_pos)
        role = Role.builtin if word.lower() in ALL_BUILTIN else None
        token = Token(word, loc=loc, role=role)
        delim = self._scan_delimiter(delimiters)

        if upper and not (word.startswith('"') or word.startswith("'")):
            token = Token(word.upper(), loc=loc)

        return token, delim


def tokenize(contents: str, filename: pathlib.Path | str = "unset") -> list[Block]:
    tok = Tokenizer(contents, filename=filename)
    return [blk.stack() for blk in tok.split_blocks()]



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from .types import Delimiter


class EndOfLine(Exception):
    pass


class ParsingError(Exception):
    pass


class UnexpectedCallName(ParsingError):
    pass


class UnexpectedAssignment(ParsingError):
    pass


class MismatchedDelimiter(ParsingError):
    open: Delimiter
    close: Delimiter

    def __init__(self, msg: str, open: Delimiter, close: Delimiter) -> None:
        super().__init__(msg)
        self.open = open
        self.close = close


class UnterminatedString(ParsingError):
    delim: Delimiter

    def __init__(self, msg: str, delim: Delimiter) -> None:
        super().__init__(msg)
        self.delim = delim


class ExtraCloseDelimiter(ParsingError):
    delim: Delimiter

    def __init__(self, msg: str, delim: Delimiter) -> None:
        super().__init__(msg)
        self.delim = delim


class MissingCloseDelimiter(ParsingError):
    delims: list[Delimiter]

    def __init__(self, msg: str, delims: list[Delimiter]) -> None:
        super().__init__(msg)
        self.delims = list(delims)


class ParticleSpeciesError(ParsingError):
    pass


class InvalidChargeError(ParticleSpeciesError):
    pass



from __future__ import annotations

from dataclasses import dataclass

from .statements import Simple, Statement
from .token import Token


@dataclass()
class Lint:
    statement: Statement
    message: str
    relevant_tokens: list[Token] | None

    def to_user_message(self):
        clsname = type(self.statement).__name__
        obj_name = str(getattr(self.statement, "name", "unnamed"))
        parts = [f"{obj_name!r} Statement of type {clsname!r}: {self.message}"]

        if self.relevant_tokens:
            parts.append("\n    Found near:")
            for tok in self.relevant_tokens:
                if tok.loc:
                    parts.append(f"{tok.quoted()} at {tok.loc}")
                else:
                    parts.append(f"{tok.quoted()}")
        return " ".join(parts)


def lint_statement(st: Statement) -> list[Lint]:
    if isinstance(st, Simple):
        if not Simple.is_known_statement(st.statement):
            return [
                Lint(
                    statement=st,
                    message=f"Statement type is unknown; this may indicate an error in parsing: {st.statement}",
                    relevant_tokens=[st.statement],
                )
            ]
    return []



"""
`latform` - a Bmad lattice parser/formatter tool.
"""

from __future__ import annotations

import argparse
import difflib
import logging
import pathlib
import sys

import rich

from . import output as output_mod
from .lint import lint_statement
from .output import format_statements
from .parser import parse, parse_file_recursive
from .statements import Statement
from .tokenizer import Tokenizer
from .types import FormatOptions, NameCase

DESCRIPTION = __doc__
logger = logging.getLogger(__name__)


def load_renames(
    rename_file: pathlib.Path | str | None,
    raw_renames: list[str] | None,
    renames: dict[str, str] | None,
):
    res = {}

    lines = []
    if rename_file:
        # todo: csv reader, maybe
        lines.extend(
            [line.split(",") for line in pathlib.Path(rename_file).read_text().splitlines()]
        )

    if raw_renames:
        lines.extend([line.split(",") for line in raw_renames])

    for from_, to in lines:
        res[from_.strip()] = to.strip()

    if renames:
        res.update(renames)

    return res


def get_diff(
    original: str, formatted: str, fromfile: pathlib.Path | str, tofile: pathlib.Path | str
):
    udiff = difflib.unified_diff(
        original.splitlines(keepends=True),
        formatted.splitlines(keepends=True),
        fromfile=str(fromfile),
        tofile=str(tofile),
    )
    return "".join(udiff)


def process_file(
    contents: str,
    filename: str | pathlib.Path,
    verbose: int = 0,
) -> list[Statement]:
    if verbose <= 0:
        return list(parse(contents, filename))

    tok = Tokenizer(contents, filename=filename)
    blocks = tok.split_blocks()
    stacked = [block.stack() for block in blocks]
    statements = []
    for idx, block in enumerate(stacked):
        # Level 1: show block header, print statement
        # Level 2: show block header, block, print statement
        # Level 3: show block header, original source, block, print statement
        if idx > 0:
            rich.print()
        rich.print(f"-- Block {idx} ({block.loc})", file=sys.stderr)
        if verbose >= 3:
            rich.print("Original source:", file=sys.stderr)
            rich.print("```", file=sys.stderr)
            rich.print(block.loc.get_string(contents), file=sys.stderr)
            rich.print("```", file=sys.stderr)
        if verbose >= 2:
            rich.print(block, file=sys.stderr)
        statement = block.parse()
        statements.append(statement)
        rich.print(statement, file=sys.stderr)
    return statements


def main(
    filename: str | pathlib.Path,
    verbose: int = 0,
    line_length: int = 100,
    max_line_length: int | None = 0,
    compact: bool = False,
    recursive: bool = False,
    in_place: bool = False,
    name_case: NameCase = "same",
    attribute_case: NameCase = "same",
    kind_case: NameCase = "same",
    builtin_case: NameCase = "same",
    section_break_character: str = "-",
    section_break_width: int = 0,
    output: pathlib.Path | str | None = None,
    diff: bool = False,
    rename_file: pathlib.Path | str | None = None,
    raw_renames: list[str] | None = None,
    renames: dict[str, str] | None = None,
) -> None:
    if str(filename) == "-":
        contents = sys.stdin.read()
        filename = "<stdin>"
        is_stdin = True
    else:
        filename = pathlib.Path(filename)
        contents = filename.read_text()
        is_stdin = False

    if verbose >= 4:
        output_mod.LATFORM_OUTPUT_DEBUG = True
        logger = logging.getLogger("latform")
        logger.setLevel("DEBUG")

    renames = load_renames(rename_file, raw_renames, renames)

    options = FormatOptions(
        line_length=line_length,
        max_line_length=max_line_length or int(line_length * 1.3),
        compact=compact,
        indent_size=2,
        indent_char=" ",
        comment_col=40,
        newline_before_new_type=not compact,
        name_case=name_case,
        attribute_case=attribute_case,
        kind_case=kind_case,
        builtin_case=builtin_case,
        section_break_character=section_break_character,
        section_break_width=section_break_width,
        renames=renames,
    )
    if recursive:
        if is_stdin:
            raise NotImplementedError(
                "Recursive parsing using a lattice from stdin is not yet supported"
            )

        files = parse_file_recursive(filename)

        # TODO this should be separate from format, I think
        # Either that or lint can be treated as an error to not reformat?
        for fn, statements in files.by_filename.items():
            for st in statements:
                for lint in lint_statement(st):
                    logger.warning(lint.to_user_message())

        if verbose > 1:
            # NOTE: double-parsing, erroneously
            for fn in files.by_filename:
                process_file(contents=fn.read_text(), filename=fn, verbose=verbose)

        if in_place:
            if diff:
                raise NotImplementedError("In-place diff is not supported (or sensible?)")
            files.reformat(options)
        else:
            for fn, statements in files.by_filename.items():
                formatted = format_statements(statements, options)
                if diff:
                    original = fn.read_text()
                    to_write = get_diff(original, formatted, fromfile=fn, tofile=fn)
                else:
                    print(f"! {fn}")
                    to_write = formatted

                print(to_write)
        return

    statements = process_file(contents=contents, filename=filename, verbose=verbose)

    for st in statements:
        for lint in lint_statement(st):
            logger.warning(lint.to_user_message())

    formatted = format_statements(statements, options=options)
    if output:
        dest_fn = output
    elif in_place and not is_stdin:
        dest_fn = filename
    else:
        dest_fn = None

    if diff:
        to_write = get_diff(contents, formatted, fromfile=filename, tofile=filename)
    else:
        to_write = formatted

    if dest_fn:
        pathlib.Path(dest_fn).write_text(to_write)
    else:
        print(to_write)


def _build_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="latform",
        description=DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    from ._version import __version__ as package_version

    parser.add_argument(
        "filename",
        help="Filename to parse (use '-' for stdin/standard input)",
    )

    parser.add_argument(
        "--rename",
        "-R",
        type=str,
        action="append",
        dest="raw_renames",
        help="Rename an element. In the form: 'old,new' (comma-delimited)",
    )

    parser.add_argument(
        "--rename-file",
        type=str,
        help="Load renames from a file. Each line should be comma-delimited in the form of `--rename`.",
    )

    parser.add_argument(
        "--diff",
        action="store_true",
        default=False,
        help="Show diff instead of formatted output",
    )

    parser.add_argument(
        "--compact",
        action="store_true",
        default=False,
        help="Compact output mode",
    )

    parser.add_argument(
        "--in-place",
        "-i",
        action="store_true",
        help="Overwrite file(s) with formatted output instead of printing to standard output",
    )
    parser.add_argument(
        "--output",
        "-o",
        action="store_true",
        help="Write to this filename (or directory, if multiple files)",
    )

    parser.add_argument(
        "--name-case",
        "--name",
        choices=("upper", "lower", "same"),
        default="upper",
        help="Case for element names, kinds, and functions",
    )

    parser.add_argument(
        "--kind-case",
        "--kind",
        choices=("upper", "lower", "same"),
        default="lower",
        help="Case for kinds (keywords)",
    )

    parser.add_argument(
        "--builtin-case",
        choices=("upper", "lower", "same"),
        default="lower",
        help="Case for builtin functions",
    )

    parser.add_argument(
        "--line-length",
        "-l",
        type=int,
        default=100,
        help="Desired line length. Some lines may exceed this (see also --max-line-length).",
    )

    parser.add_argument(
        "--max-line-length",
        "-m",
        type=int,
        default=None,
        help="Force lines over this length to be multilined. Defaults to 130%% of line_length.",
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="Increase debug verbosity",
    )

    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=package_version,
        help="Show the latform version number and exit.",
    )

    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="Recursively (-r) parse lattice files, following call statements",
    )

    parser.add_argument(
        "--section-break-character",
        type=str,
        default="-",
        help="Section break character.  By default --line-length characters, unless overridden by --section-break-width",
    )

    parser.add_argument(
        "--section-break-width",
        type=int,
        default=None,
        help="Section break line width.  By default --line-length characters",
    )

    parser.add_argument(
        "--log",
        "-L",
        dest="log_level",
        default="INFO",
        choices=("DEBUG", "INFO", "WARNING", "CRITICAL"),
        help="Python logging level (e.g. DEBUG, INFO, WARNING)",
    )

    return parser


def cli_main(args: list[str] | None = None) -> None:
    """
    CLI entrypoint main.

    Parameters
    ----------
    args : list of str, optional
        Command-line arguments to parse and pass to :func:`main()`.
    """
    parsed = _build_argparser().parse_args(args=args)
    kwargs = vars(parsed)
    log_level = kwargs.pop("log_level")

    # Adjust the package-level logger level as requested:
    logger = logging.getLogger("latform")
    logger.setLevel(log_level)
    logging.basicConfig()
    return main(**kwargs)


if __name__ == "__main__":
    cli_main()



from __future__ import annotations

import logging
import pathlib
from dataclasses import dataclass, field
from typing import Sequence

from .const import EQUALS
from .exceptions import UnexpectedAssignment
from .statements import (
    Assignment,
    Constant,
    Element,
    ElementList,
    Empty,
    Line,
    NonstandardParameter,
    Parameter,
    Simple,
    Statement,
    get_call_filename,
)
from .token import Comments, Role, Token
from .tokenizer import tokenize
from .types import (
    COMMA,
    Attribute,
    Block,
    CallName,
    Delimiter,
    FormatOptions,
    Seq,
    TokenizerItem,
)
from .util import partition_items

logger = logging.getLogger(__name__)


def _make_attribute(item: Attribute | Token | Seq) -> Attribute:
    if isinstance(item, Attribute):
        return item
    if isinstance(item, Delimiter):
        raise ValueError(f"Unexpected delimiter found in place of attribute: {item} at {item.loc}")
    if isinstance(item, Token):
        return Attribute(name=item)
    if isinstance(item, Seq):
        return Attribute(name=item)
    raise ValueError(f"Unexpected item found in place of attribute: {item} at {item.loc}")


def _make_attribute_list(items: list[TokenizerItem]) -> list[Attribute]:
    item = Seq.from_items(items)
    if not isinstance(item, Seq):
        return [_make_attribute(item)]

    return [_make_attribute(item) for item in item.items]


def _is_multipass_marker(blk: Block) -> bool:
    """Check if Array represents a multipass marker."""
    return (
        blk.opener == "["
        and len(blk.items) == 1
        and isinstance(blk.items[0], Token)
        and str(blk.items[0]).lower() == "multipass"
    )


def _extract_leading_comment(first: TokenizerItem) -> Comments:
    comments = first.comments.clone()
    first.comments.clear()
    return comments


def _nab_comments(items) -> Comments:
    res = Comments()
    inline = []
    for item in items:
        if isinstance(item, Seq):
            comment = _nab_comments(item.items)
        else:
            comment = item.comments

        res.pre.extend(comment.pre)
        comment.pre.clear()
        if comment.inline:
            inline.append(comment.inline)
            comment.inline = None

    if not inline:
        pass
    elif len(inline) == 1:
        res.inline = inline[0]
    else:
        res.pre.extend(inline[:-1])
        res.inline = inline[-1]
    return res


def _line_elements_from_block(block: Block) -> Seq:
    if block.opener != "(":
        raise ValueError(f"Unexpected block opener: {block.opener}")

    eles = Seq.from_delimited_block(block, delimiter=COMMA)
    assert isinstance(eles, Seq)
    for ele in eles.items:
        match ele:
            case Seq(items=["-", "-", name]):
                # Element reversal and reflection
                ele.items = [Delimiter("--"), name.with_(role=Role.name_)]
            case Token():
                ele.role = Role.name_
            case Seq():
                ele.items = ele.with_(role=Role.name_).items
            case _:
                raise ValueError(f"Unexpected type found in element list: {type(ele)} {ele=}")

    return eles


known_parameters_keyed = {(param.target, param.name): param for param in Parameter.known}


def fix_parameter_value(
    target: Token, name: Token, value: Token | Seq, raw_value: list[Token | Block]
):
    key = (str(target).lower(), str(name).lower())
    try:
        param = known_parameters_keyed[key]
    except KeyError:
        return value

    if param.type == "species":
        if isinstance(value, Seq):
            return value.to_token(include_opener=False).replace(" ", "")
        return value

    if param.type is str:
        value = Token.join(raw_value)
        if not value.is_quoted_string:
            return Token(f'"{value}"', loc=value.loc, comments=value.comments)
        return value

    if param.type == "geometry":
        if isinstance(value, Token):
            if value.lower().startswith("o"):
                return Token("open", loc=value.loc, comments=value.comments)
            if value.lower().startswith("c"):
                return Token("closed", loc=value.loc, comments=value.comments)
        return value

    if param.type is bool:
        if isinstance(value, Token):
            if value.lower().startswith("t"):
                return Token("True", loc=value.loc, comments=value.comments)
            if value.lower().startswith("f"):
                return Token("False", loc=value.loc, comments=value.comments)
        return value

    return value


def parse_items(items: list[TokenizerItem]):
    if not items:
        raise ValueError("No items provided")

    first = items[0]
    comments = _extract_leading_comment(first)
    first.comments.clear()

    match items:
        # These two cases are handled at the end, along with general 'parameters'.
        # case [Token("beginning") as target, Block() as name, "=", _ as value]:
        # case [Token("parameter") as target, Block() as name, "=", Token() as value]:

        case [Token("redef"), ":", Token() as name, "=", *rest]:
            value = Seq.from_items(rest)
            if isinstance(value, Attribute):
                raise UnexpectedAssignment(
                    f"Unexpected named attribute assignment: {value} at {value.loc}"
                )
            return Constant(
                comments=comments,
                name=name.with_(role=Role.name_),
                value=value,
                redef=True,
            )

        case [Token() as name, "=", *rest]:
            value = Seq.from_items(rest)
            if isinstance(value, Attribute):
                raise UnexpectedAssignment(
                    f"Unexpected named attribute assignment: {value} at {value.loc}"
                )
            return Constant(comments=comments, name=name.with_(role=Role.name_), value=value)

        case [Token() as name, ":", Token("list"), "=", Block(opener="(") as elements_block]:
            return ElementList(
                comments=comments,
                name=name.with_(role=Role.name_),
                elements=_line_elements_from_block(elements_block),
            )

        case [Token() as name, ":", Token("line"), "=", Block(opener="(") as elements_block]:
            return Line(
                comments=comments,
                name=name.with_(role=Role.name_),
                elements=_line_elements_from_block(elements_block),
            )

        case [
            Token() as name,
            ":",
            Token("line"),
            Block(opener="[") as multipass,
            "=",
            Block(opener="(") as elements_block,
        ] if _is_multipass_marker(multipass):
            return Line(
                comments=comments,
                name=name.with_(role=Role.name_),
                elements=_line_elements_from_block(elements_block),
                multipass=True,
            )

        case [
            Token() as name,
            Block(opener="(") as line_args,
            ":",
            Token("line"),
            "=",
            Block(opener="(") as elements_block,
        ]:
            assert isinstance(name, Token)
            return Line(
                comments=comments,
                name=CallName(
                    name=name.with_(role=Role.name_),
                    args=Seq.from_item(line_args),
                ),
                elements=_line_elements_from_block(elements_block),
            )

        case [Token() as name, ":", Token() as element_type, *rest]:
            match rest:
                case ["=", Block(opener="{") as ele_list, *after]:
                    if after and after[0] == COMMA:
                        after = after[1:]
                    return Element(
                        comments=comments,
                        name=name.with_(role=Role.name_),
                        keyword=element_type,
                        ele_list=Seq.from_delimited_block(ele_list, delimiter=COMMA),
                        attributes=_make_attribute_list(after),
                    )

                case [",", *after]:
                    return Element(
                        comments=comments,
                        name=name.with_(role=Role.name_),
                        keyword=element_type,
                        attributes=_make_attribute_list(after),
                    )

                case []:
                    return Element(
                        comments=comments,
                        name=name.with_(role=Role.name_),
                        keyword=element_type,
                        attributes=[],
                    )

        case [Token() as stmt]:
            if stmt == "":
                return Empty(comments=comments)

            return Simple(comments=comments, statement=stmt, arguments=[])

    if isinstance(first, Token):
        if first.lower() in {"print", "parser_debug"}:
            args = items[1:]
            if args[0] == COMMA:
                args = args[1:]
            return Simple(
                comments=comments,
                statement=first,
                arguments=[item.to_token() if isinstance(item, Block) else item for item in args],
            )

        if Simple.is_known_statement(first):
            args = items[1:]

            if args[0] == COMMA:
                args = args[1:]

            attrs = _make_attribute_list(args)
            assert isinstance(first, Token)

            return Simple(
                comments=comments,
                statement=first,
                arguments=attrs,
            )

    # Match assignment patterns
    if EQUALS in items:
        before_equals, _, after_equals = partition_items(items, EQUALS)
        if not before_equals or not after_equals:
            raise ValueError("Unhandled assignment: missing name or value")

        value = Seq.from_items(after_equals)

        if isinstance(value, Attribute):
            raise UnexpectedAssignment(
                f"Unexpected named attribute assignment: {value} at {value.loc}"
            )

        match before_equals:
            # Parameter with [attribute] syntax: target[name] = value
            case [*target, Block(opener="[") as name_block]:
                target = Seq.from_items(target)

                # This couldn't be an attribute as there's no '=' in there
                assert not isinstance(target, Attribute)

                try:
                    name = name_block.squeeze_single_token()
                    if "%" in name:
                        raise ValueError("Nonstandard parameter name")
                except ValueError:
                    name = name_block.to_token(include_opener=False)
                    name = Token(name.replace(" ", ""), comments=name.comments, loc=name.loc)
                    cls = NonstandardParameter
                else:
                    cls = Parameter

                if isinstance(target, Token):
                    value = fix_parameter_value(target, name, value, raw_value=after_equals)

                return cls(
                    comments=comments,
                    target=target,
                    name=name.with_(role=Role.name_),
                    value=value,
                )
            # Generic assignment: name = value
            case _:
                name = Seq.from_items(before_equals)
                # This couldn't be an attribute as there's no '=' in there
                assert not isinstance(name, Attribute)
                return Assignment(
                    name=name.with_(role=Role.name_),
                    value=value,
                    comments=comments,
                )

    raise ValueError("Unhandled - unknown")


def get_named_items(statements: Sequence[Statement]) -> dict[Token, Statement]:
    named_items = {}
    for statement in statements:
        if isinstance(statement, (Element, Constant)):
            named_items[statement.name.upper()] = statement
        elif isinstance(statement, Line):
            if isinstance(statement.name, CallName):
                named_items[statement.name.name.upper()] = statement
            else:
                named_items[statement.name.upper()] = statement
    return named_items


def parse(
    contents: str, filename: pathlib.Path | str = "unset", annotate: bool = True
) -> Sequence[Statement]:
    blocks = tokenize(contents, filename)
    res = [block.parse() for block in blocks]
    if annotate:
        named = get_named_items(res)
        for st in res:
            st.annotate(named=named)

    return res


def parse_file(filename: pathlib.Path | str, annotate: bool = True) -> Sequence[Statement]:
    contents = pathlib.Path(filename).read_text()
    return parse(contents=contents, filename=filename, annotate=annotate)


def parse_file_recursive(filename: pathlib.Path | str, annotate: bool = True) -> Files:
    files = Files(main=pathlib.Path(filename))
    files.parse()
    if annotate:
        files.annotate()
    return files


@dataclass
class Files:
    main: pathlib.Path
    stack: list[tuple[pathlib.Path, pathlib.Path]] = field(default_factory=list)
    by_filename: dict[pathlib.Path, list[Statement]] = field(default_factory=dict)
    local_file_to_source_filename: dict[pathlib.Path, str] = field(default_factory=dict)
    filename_calls: dict[pathlib.Path, list[pathlib.Path]] = field(default_factory=dict)

    def _add_file_by_statement(self, statement_filename: pathlib.Path, st: Simple):
        assert isinstance(st, Simple) and st.statement == "call"
        sub_filename, fn = get_call_filename(
            st, caller_directory=statement_filename.parent, expand_vars=True
        )
        self.local_file_to_source_filename[fn] = sub_filename
        logger.debug(f"Adding {sub_filename} relative to {statement_filename.parent} which is {fn}")
        self.stack.append((fn, statement_filename.parent))
        self.filename_calls.setdefault(statement_filename, [])
        self.filename_calls[statement_filename].append(fn)

    @property
    def call_graph_edges(self):
        graph = []
        for path_fn, calls in self.filename_calls.items():
            fn = self.local_file_to_source_filename[path_fn]
            for call_path in calls:
                call_fn = self.local_file_to_source_filename[call_path]
                graph.append((fn, call_fn))
        return graph

    def parse(self):
        main = self.main.resolve()

        if not self.stack:
            self.stack = [(main, main.parent)]
            self.local_file_to_source_filename[main] = main.name

        while self.stack:
            filename, parent_dir = self.stack.pop()

            filename = parent_dir / filename
            statements = list(parse_file(filename, annotate=False))
            self.by_filename[filename] = statements
            for st in statements:
                if isinstance(st, Simple) and st.statement == "call":
                    self._add_file_by_statement(statement_filename=filename, st=st)

        return self.by_filename

    def annotate(self):
        named = self.get_named_items()
        for statements in self.by_filename.values():
            for st in statements:
                st.annotate(named=named)

    def get_named_items(self) -> dict[Token, Statement]:
        named_items = {}
        for statements in self.by_filename.values():
            new_items = get_named_items(statements)
            # TODO: potential for linting with redef
            named_items.update(new_items)
        return named_items

    def reformat(self, options: FormatOptions, *, dest: pathlib.Path | None = None) -> None:
        from .output import format_statements

        for fn, statements in self.by_filename.items():
            formatted = format_statements(statements, options)
            fn.write_text(formatted)
